#!/usr/bin/python

# Copyright (c) 2016 Michael Gruener <michael.gruener@chaosmoon.net>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: cloudflare_dns
author:
  - Michael Gruener (@mgruener)
short_description: Manage Cloudflare DNS records
description:
  - 'Manages DNS records using the Cloudflare API, see the docs: U(https://api.cloudflare.com/).'
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  api_token:
    description:
      - API token.
      - Required for API token authentication.
      - "You can obtain your API token from the bottom of the Cloudflare 'My Account' page, found here: U(https://dash.cloudflare.com/)."
      - Can be specified in E(CLOUDFLARE_TOKEN) environment variable since community.general 2.0.0.
    type: str
    version_added: '0.2.0'
  account_api_key:
    description:
      - Account API key.
      - Required for API keys authentication.
      - "You can obtain your API key from the bottom of the Cloudflare 'My Account' page, found here: U(https://dash.cloudflare.com/)."
    type: str
    aliases: [account_api_token]
  account_email:
    description:
      - Account email. Required for API keys authentication.
    type: str
  algorithm:
    description:
      - Algorithm number.
      - Required for O(type=DS) and O(type=SSHFP) when O(state=present).
    type: int
  cert_usage:
    description:
      - Certificate usage number.
      - Required for O(type=TLSA) when O(state=present).
    type: int
    choices: [0, 1, 2, 3]
  comment:
    description:
      - Comments or notes about the DNS record.
    type: str
    version_added: 10.1.0
  flag:
    description:
      - Issuer Critical Flag.
      - Should be set either to V(0) for not critical, or to V(128) for critical.
      - Before community.general 12.3.0, only V(0) and V(1) were valid options.
        The value V(1) is still available for backwards-compatibility reasons.
      - Required for O(type=CAA) when O(state=present).
    type: int
    choices: [0, 1, 128]
    version_added: 8.0.0
  tag:
    description:
      - CAA issue restriction.
      - Required for O(type=CAA) when O(state=present).
    type: str
    choices: [issue, issuewild, iodef]
    version_added: 8.0.0
  hash_type:
    description:
      - Hash type number.
      - Required for O(type=DS), O(type=SSHFP) and O(type=TLSA) when O(state=present).
    type: int
    choices: [1, 2]
  key_tag:
    description:
      - DNSSEC key tag.
      - Needed for O(type=DS) when O(state=present).
    type: int
  port:
    description:
      - Service port.
      - Required for O(type=SRV) and O(type=TLSA).
    type: int
  priority:
    description:
      - Record priority.
      - Required for O(type=MX) and O(type=SRV).
    default: 1
    type: int
  proto:
    description:
      - Service protocol. Required for O(type=SRV) and O(type=TLSA).
      - Common values are TCP and UDP.
    type: str
  proxied:
    description:
      - Proxy through Cloudflare network or just use DNS.
    type: bool
    default: false
  record:
    description:
      - Record to add.
      - Required if O(state=present).
      - Default is V(@) (that is, the zone name).
    type: str
    default: '@'
    aliases: [name]
  selector:
    description:
      - Selector number.
      - Required for O(type=TLSA) when O(state=present).
    choices: [0, 1]
    type: int
  service:
    description:
      - Record service.
      - Required for O(type=SRV).
    type: str
  solo:
    description:
      - Whether the record should be the only one for that record type and record name.
      - Only use with O(state=present).
      - This deletes all other records with the same record name and type.
    type: bool
  state:
    description:
      - Whether the record(s) should exist or not.
    type: str
    choices: [absent, present]
    default: present
  tags:
    description:
      - Custom tags for the DNS record.
    type: list
    elements: str
    version_added: 10.1.0
  timeout:
    description:
      - Timeout for Cloudflare API calls.
    type: int
    default: 30
  ttl:
    description:
      - The TTL to give the new record.
      - Must be between V(120) and V(2,147,483,647) seconds, or V(1) for automatic.
    type: int
    default: 1
  type:
    description:
      - The type of DNS record to create. Required if O(state=present).
      - Support for V(SPF) has been removed from community.general 9.0.0 since that record type is no longer supported by
        CloudFlare.
      - Support for V(PTR) has been added in community.general 11.1.0.
    type: str
    choices: [A, AAAA, CNAME, DS, MX, NS, SRV, SSHFP, TLSA, CAA, TXT, PTR]
  value:
    description:
      - The record value.
      - Required for O(state=present).
    type: str
    aliases: [content]
  weight:
    description:
      - Service weight.
      - Required for O(type=SRV).
    type: int
    default: 1
  zone:
    description:
      - The name of the Zone to work with (for example V(example.com)).
      - The Zone must already exist.
    type: str
    required: true
    aliases: [domain]
"""

EXAMPLES = r"""
- name: Create a test.example.net A record to point to 127.0.0.1
  community.general.cloudflare_dns:
    zone: example.net
    record: test
    type: A
    value: 127.0.0.1
    account_email: test@example.com
    account_api_key: dummyapitoken
  register: record

- name: Create a record using api token
  community.general.cloudflare_dns:
    zone: example.net
    record: test
    type: A
    value: 127.0.0.1
    api_token: dummyapitoken

- name: Create a record with comment and tags
  community.general.cloudflare_dns:
    zone: example.net
    record: test
    type: A
    value: 127.0.0.1
    comment: Local test website
    tags:
      - test
      - local
    api_token: dummyapitoken

- name: Create a example.net CNAME record to example.com
  community.general.cloudflare_dns:
    zone: example.net
    type: CNAME
    value: example.com
    account_email: test@example.com
    account_api_key: dummyapitoken
    state: present

- name: Change its TTL
  community.general.cloudflare_dns:
    zone: example.net
    type: CNAME
    value: example.com
    ttl: 600
    account_email: test@example.com
    account_api_key: dummyapitoken
    state: present

- name: Delete the record
  community.general.cloudflare_dns:
    zone: example.net
    type: CNAME
    value: example.com
    account_email: test@example.com
    account_api_key: dummyapitoken
    state: absent

- name: Create a example.net CNAME record to example.com and proxy through Cloudflare's network
  community.general.cloudflare_dns:
    zone: example.net
    type: CNAME
    value: example.com
    proxied: true
    account_email: test@example.com
    account_api_key: dummyapitoken
    state: present

# This deletes all other TXT records named "test.example.net"
- name: Create TXT record "test.example.net" with value "unique value"
  community.general.cloudflare_dns:
    domain: example.net
    record: test
    type: TXT
    value: unique value
    solo: true
    account_email: test@example.com
    account_api_key: dummyapitoken
    state: present

- name: Create an SRV record _foo._tcp.example.net
  community.general.cloudflare_dns:
    domain: example.net
    service: foo
    proto: tcp
    port: 3500
    priority: 10
    weight: 20
    type: SRV
    value: fooserver.example.net

- name: Create a SSHFP record login.example.com
  community.general.cloudflare_dns:
    zone: example.com
    record: login
    type: SSHFP
    algorithm: 4
    hash_type: 2
    value: 9dc1d6742696d2f51ca1f1a78b3d16a840f7d111eb9454239e70db31363f33e1

- name: Create a TLSA record _25._tcp.mail.example.com
  community.general.cloudflare_dns:
    zone: example.com
    record: mail
    port: 25
    proto: tcp
    type: TLSA
    cert_usage: 3
    selector: 1
    hash_type: 1
    value: 6b76d034492b493e15a7376fccd08e63befdad0edab8e442562f532338364bf3

- name: Create a CAA record subdomain.example.com
  community.general.cloudflare_dns:
    zone: example.com
    record: subdomain
    type: CAA
    flag: 0
    tag: issue
    value: ca.example.com

- name: Create a DS record for subdomain.example.com
  community.general.cloudflare_dns:
    zone: example.com
    record: subdomain
    type: DS
    key_tag: 5464
    algorithm: 8
    hash_type: 2
    value: B4EB5AC4467D2DFB3BAF9FB9961DC1B6FED54A58CDFAA3E465081EC86F89BFAB

- name: Create PTR record "1.2.0.192.in-addr.arpa" with value "test.example.com"
  community.general.cloudflare_dns:
    zone: 2.0.192.in-addr.arpa
    record: 1
    type: PTR
    value: test.example.com
    state: present
"""

RETURN = r"""
record:
  description: A dictionary containing the record data.
  returned: success, except on record deletion
  type: dict
  contains:
    comment:
      description: Comments or notes about the DNS record.
      returned: success
      type: str
      sample: Domain verification record
      version_added: 10.1.0
    comment_modified_on:
      description: When the record comment was last modified. Omitted if there is no comment.
      returned: success
      type: str
      sample: "2024-01-01T05:20:00.12345Z"
      version_added: 10.1.0
    content:
      description: The record content (details depend on record type).
      returned: success
      type: str
      sample: 192.0.2.91
    created_on:
      description: The record creation date.
      returned: success
      type: str
      sample: "2016-03-25T19:09:42.516553Z"
    data:
      description: Additional record data.
      returned: success, if type is SRV, DS, SSHFP TLSA or CAA
      type: dict
      sample:
        {
          "name": "jabber",
          "port": 8080,
          "priority": 10,
          "proto": "_tcp",
          "service": "_xmpp",
          "target": "jabberhost.sample.com",
          "weight": 5
        }
    id:
      description: The record ID.
      returned: success
      type: str
      sample: f9efb0549e96abcb750de63b38c9576e
    locked:
      description: No documentation available.
      returned: success
      type: bool
      sample: false
    meta:
      description: Extra Cloudflare-specific information about the record.
      returned: success
      type: dict
      sample: {"auto_added": false}
    modified_on:
      description: Record modification date.
      returned: success
      type: str
      sample: "2016-03-25T19:09:42.516553Z"
    name:
      description: The record name as FQDN (including _service and _proto for SRV).
      returned: success
      type: str
      sample: www.sample.com
    priority:
      description: Priority of the MX record.
      returned: success, if type is MX
      type: int
      sample: 10
    proxiable:
      description: Whether this record can be proxied through Cloudflare.
      returned: success
      type: bool
      sample: false
    proxied:
      description: Whether the record is proxied through Cloudflare.
      returned: success
      type: bool
      sample: false
    tags:
      description: Custom tags for the DNS record.
      returned: success
      type: list
      elements: str
      sample: ["production", "app"]
      version_added: 10.1.0
    tags_modified_on:
      description: When the record tags were last modified. Omitted if there are no tags.
      returned: success
      type: str
      sample: "2025-01-01T05:20:00.12345Z"
      version_added: 10.1.0
    ttl:
      description: The time-to-live for the record.
      returned: success
      type: int
      sample: 300
    type:
      description: The record type.
      returned: success
      type: str
      sample: A
    zone_id:
      description: The ID of the zone containing the record.
      returned: success
      type: str
      sample: abcede0bf9f0066f94029d2e6b73856a
    zone_name:
      description: The name of the zone containing the record.
      returned: success
      type: str
      sample: sample.com
"""

import json
from urllib.parse import urlencode

from ansible.module_utils.basic import AnsibleModule, env_fallback
from ansible.module_utils.common.text.converters import to_text
from ansible.module_utils.urls import fetch_url


def lowercase_string(param):
    return param.lower() if isinstance(param, str) else param


def join_str(sep, *args):
    return sep.join([str(arg) for arg in args])


class CloudflareAPI:
    cf_api_endpoint = "https://api.cloudflare.com/client/v4"
    changed = False

    def __init__(self, module):
        self.module = module
        self.api_token = module.params["api_token"]
        self.account_api_key = module.params["account_api_key"]
        self.account_email = module.params["account_email"]
        self.algorithm = module.params["algorithm"]
        self.cert_usage = module.params["cert_usage"]
        self.comment = module.params["comment"]
        self.hash_type = module.params["hash_type"]
        self.flag = module.params["flag"]
        self.tag = module.params["tag"]
        self.tags = module.params["tags"]
        self.key_tag = module.params["key_tag"]
        self.port = module.params["port"]
        self.priority = module.params["priority"]
        self.proto = lowercase_string(module.params["proto"])
        self.proxied = module.params["proxied"]
        self.selector = module.params["selector"]
        self.record = lowercase_string(module.params["record"])
        self.service = lowercase_string(module.params["service"])
        self.is_solo = module.params["solo"]
        self.state = module.params["state"]
        self.timeout = module.params["timeout"]
        self.ttl = module.params["ttl"]
        self.type = module.params["type"]
        self.value = module.params["value"]
        self.weight = module.params["weight"]
        self.zone = lowercase_string(module.params["zone"])

        if self.record == "@":
            self.record = self.zone

        if (self.type in ["CNAME", "NS", "MX", "SRV"]) and (self.value is not None):
            self.value = self.value.rstrip(".").lower()

        if (self.type == "AAAA") and (self.value is not None):
            self.value = self.value.lower()

        if self.type == "SRV":
            if (self.proto is not None) and (not self.proto.startswith("_")):
                self.proto = f"_{self.proto}"
            if (self.service is not None) and (not self.service.startswith("_")):
                self.service = f"_{self.service}"

        if self.type == "TLSA":
            if (self.proto is not None) and (not self.proto.startswith("_")):
                self.proto = f"_{self.proto}"
            if self.port is not None:
                self.port = f"_{self.port}"

        if not self.record.endswith(self.zone):
            self.record = join_str(".", self.record, self.zone)

        if self.type == "DS":
            if self.record == self.zone:
                self.module.fail_json(msg="DS records only apply to subdomains.")

    def _cf_simple_api_call(self, api_call, method="GET", payload=None):
        if self.api_token:
            headers = {
                "Authorization": f"Bearer {self.api_token}",
                "Content-Type": "application/json",
            }
        else:
            headers = {
                "X-Auth-Email": self.account_email,
                "X-Auth-Key": self.account_api_key,
                "Content-Type": "application/json",
            }
        data = None
        if payload:
            try:
                data = json.dumps(payload)
            except Exception as e:
                self.module.fail_json(msg=f"Failed to encode payload as JSON: {e} ")

        resp, info = fetch_url(
            self.module,
            self.cf_api_endpoint + api_call,
            headers=headers,
            data=data,
            method=method,
            timeout=self.timeout,
        )

        if info["status"] not in [200, 304, 400, 401, 403, 429, 405, 415]:
            self.module.fail_json(
                msg=f"Failed API call {api_call}; got unexpected HTTP code {info['status']}: {info.get('msg')}"
            )

        error_msg = ""
        if info["status"] == 401:
            # Unauthorized
            error_msg = (
                f"API user does not have permission; Status: {info['status']}; Method: {method}: Call: {api_call}"
            )
        elif info["status"] == 403:
            # Forbidden
            error_msg = f"API request not authenticated; Status: {info['status']}; Method: {method}: Call: {api_call}"
        elif info["status"] == 429:
            # Too many requests
            error_msg = f"API client is rate limited; Status: {info['status']}; Method: {method}: Call: {api_call}"
        elif info["status"] == 405:
            # Method not allowed
            error_msg = (
                f"API incorrect HTTP method provided; Status: {info['status']}; Method: {method}: Call: {api_call}"
            )
        elif info["status"] == 415:
            # Unsupported Media Type
            error_msg = f"API request is not valid JSON; Status: {info['status']}; Method: {method}: Call: {api_call}"
        elif info["status"] == 400:
            # Bad Request
            error_msg = f"API bad request; Status: {info['status']}; Method: {method}: Call: {api_call}"

        result = None
        try:
            content = resp.read()
        except AttributeError:
            content = None

        if not content:
            if info["body"]:
                content = info["body"]
            else:
                error_msg += "; The API response was empty"

        if content:
            try:
                result = json.loads(to_text(content, errors="surrogate_or_strict"))
            except getattr(json, "JSONDecodeError", ValueError) as e:
                error_msg += f"; Failed to parse API response with error {e}: {content}"

        # Without a valid/parsed JSON response no more error processing can be done
        if result is None:
            self.module.fail_json(msg=error_msg)

        if "success" not in result:
            error_msg += f"; Unexpected error details: {result.get('error')}"
            self.module.fail_json(msg=error_msg)

        if not result["success"]:
            error_msg += "; Error details: "
            for error in result["errors"]:
                error_msg += f"code: {error['code']}, error: {error['message']}; "
                if "error_chain" in error:
                    for chain_error in error["error_chain"]:
                        error_msg += f"code: {chain_error['code']}, error: {chain_error['message']}; "
            self.module.fail_json(msg=error_msg)

        return result, info["status"]

    def _cf_api_call(self, api_call, method="GET", payload=None):
        result, status = self._cf_simple_api_call(api_call, method, payload)

        data = result["result"]

        if "result_info" in result:
            pagination = result["result_info"]
            if pagination["total_pages"] > 1:
                next_page = int(pagination["page"]) + 1
                parameters = [f"page={next_page}"]
                # strip "page" parameter from call parameters (if there are any)
                if "?" in api_call:
                    raw_api_call, query = api_call.split("?", 1)
                    parameters += [param for param in query.split("&") if not param.startswith("page")]
                else:
                    raw_api_call = api_call
                while next_page <= pagination["total_pages"]:
                    raw_api_call += f"?{'&'.join(parameters)}"
                    result, status = self._cf_simple_api_call(raw_api_call, method, payload)
                    data += result["result"]
                    next_page += 1

        return data, status

    def _get_zone_id(self, zone=None):
        if not zone:
            zone = self.zone

        zones = self.get_zones(zone)
        if len(zones) > 1:
            self.module.fail_json(msg=f"More than one zone matches {zone}")

        if len(zones) < 1:
            self.module.fail_json(msg=f"No zone found with name {zone}")

        return zones[0]["id"]

    def get_zones(self, name=None):
        if not name:
            name = self.zone
        param = ""
        if name:
            param = f"?{urlencode({'name': name})}"
        zones, status = self._cf_api_call(f"/zones{param}")
        return zones

    def get_dns_records(self, zone_name=None, type=None, record=None, value=""):
        if not zone_name:
            zone_name = self.zone
        if not type:
            type = self.type
        if not record:
            record = self.record
        # necessary because None as value means to override user
        # set module value
        if (not value) and (value is not None):
            value = self.value

        zone_id = self._get_zone_id()
        api_call = f"/zones/{zone_id}/dns_records"
        query = {}
        if type:
            query["type"] = type
        if record:
            query["name"] = record
        if value:
            query["content"] = value
        if query:
            api_call += f"?{urlencode(query)}"

        records, status = self._cf_api_call(api_call)
        return records

    def delete_dns_records(self, solo):
        records = []
        content = self.value
        search_record = self.record
        if self.type == "SRV":
            if not (self.value is None or self.value == ""):
                content = join_str("\t", self.weight, self.port, self.value)
            search_record = join_str(".", self.service, self.proto, self.record)
        elif self.type == "DS":
            if not (self.value is None or self.value == ""):
                content = join_str("\t", self.key_tag, self.algorithm, self.hash_type, self.value)
        elif self.type == "SSHFP":
            if not (self.value is None or self.value == ""):
                content = join_str(" ", self.algorithm, self.hash_type, self.value.upper())
        elif self.type == "TLSA":
            if not (self.value is None or self.value == ""):
                content = join_str("\t", self.cert_usage, self.selector, self.hash_type, self.value)
            search_record = join_str(".", self.port, self.proto, self.record)
        if solo:
            search_value = None
        else:
            search_value = content

        zone_id = self._get_zone_id(self.zone)
        records = self.get_dns_records(self.zone, self.type, search_record, search_value)

        for rr in records:
            if solo:
                if not ((rr["type"] == self.type) and (rr["name"] == search_record) and (rr["content"] == content)):
                    self.changed = True
                    if not self.module.check_mode:
                        result, info = self._cf_api_call(f"/zones/{zone_id}/dns_records/{rr['id']}", "DELETE")
            else:
                self.changed = True
                if not self.module.check_mode:
                    result, info = self._cf_api_call(f"/zones/{zone_id}/dns_records/{rr['id']}", "DELETE")
        return self.changed

    def ensure_dns_record(self):
        search_value = self.value
        search_record = self.record
        new_record = None

        if self.type in ["A", "AAAA", "CNAME", "TXT", "MX", "NS", "PTR"]:
            if not self.value:
                self.module.fail_json(msg="You must provide a non-empty value to create this record type")

            # there can only be one CNAME per record
            # ignoring the value when searching for existing
            # CNAME records allows us to update the value if it
            # changes
            if self.type == "CNAME":
                search_value = None

            new_record = {"type": self.type, "name": self.record, "content": self.value, "ttl": self.ttl}

        if self.type in ["A", "AAAA", "CNAME"]:
            new_record["proxied"] = self.proxied

        if self.type == "MX":
            for attr in [self.priority, self.value]:
                if (attr is None) or (attr == ""):
                    self.module.fail_json(msg="You must provide priority and a value to create this record type")
            new_record = {
                "type": self.type,
                "name": self.record,
                "content": self.value,
                "priority": self.priority,
                "ttl": self.ttl,
            }

        if self.type == "SRV":
            for attr in [self.port, self.priority, self.proto, self.service, self.weight, self.value]:
                if (attr is None) or (attr == ""):
                    self.module.fail_json(
                        msg="You must provide port, priority, proto, service, weight and a value to create this record type"
                    )
            srv_data = {
                "target": self.value,
                "port": self.port,
                "weight": self.weight,
                "priority": self.priority,
            }

            new_record = {
                "type": self.type,
                "name": join_str(".", self.service, self.proto, self.record),
                "ttl": self.ttl,
                "data": srv_data,
            }
            search_value = join_str("\t", self.weight, self.port, self.value)
            search_record = join_str(".", self.service, self.proto, self.record)

        if self.type == "DS":
            for attr in [self.key_tag, self.algorithm, self.hash_type, self.value]:
                if (attr is None) or (attr == ""):
                    self.module.fail_json(
                        msg="You must provide key_tag, algorithm, hash_type and a value to create this record type"
                    )
            ds_data = {
                "key_tag": self.key_tag,
                "algorithm": self.algorithm,
                "digest_type": self.hash_type,
                "digest": self.value,
            }
            new_record = {
                "type": self.type,
                "name": self.record,
                "data": ds_data,
                "ttl": self.ttl,
            }
            search_value = join_str("\t", self.key_tag, self.algorithm, self.hash_type, self.value)

        if self.type == "SSHFP":
            for attr in [self.algorithm, self.hash_type, self.value]:
                if (attr is None) or (attr == ""):
                    self.module.fail_json(
                        msg="You must provide algorithm, hash_type and a value to create this record type"
                    )
            sshfp_data = {
                "fingerprint": self.value.upper(),
                "type": self.hash_type,
                "algorithm": self.algorithm,
            }
            new_record = {
                "type": self.type,
                "name": self.record,
                "data": sshfp_data,
                "ttl": self.ttl,
            }
            search_value = join_str(" ", self.algorithm, self.hash_type, self.value)

        if self.type == "TLSA":
            for attr in [self.port, self.proto, self.cert_usage, self.selector, self.hash_type, self.value]:
                if (attr is None) or (attr == ""):
                    self.module.fail_json(
                        msg="You must provide port, proto, cert_usage, selector, hash_type and a value to create this record type"
                    )
            search_record = join_str(".", self.port, self.proto, self.record)
            tlsa_data = {
                "usage": self.cert_usage,
                "selector": self.selector,
                "matching_type": self.hash_type,
                "certificate": self.value,
            }
            new_record = {
                "type": self.type,
                "name": search_record,
                "data": tlsa_data,
                "ttl": self.ttl,
            }
            search_value = join_str("\t", self.cert_usage, self.selector, self.hash_type, self.value)

        if self.type == "CAA":
            for attr in [self.flag, self.tag, self.value]:
                if attr == "":
                    self.module.fail_json(msg="You must provide flag, tag and a value to create this record type")
            caa_data = {
                "flags": self.flag,
                "tag": self.tag,
                "value": self.value,
            }
            new_record = {
                "type": self.type,
                "name": self.record,
                "data": caa_data,
                "ttl": self.ttl,
            }
            search_value = None

        new_record["comment"] = self.comment or None
        new_record["tags"] = self.tags or []

        zone_id = self._get_zone_id(self.zone)
        records = self.get_dns_records(self.zone, self.type, search_record, search_value)
        # in theory this should be impossible as cloudflare does not allow
        # the creation of duplicate records but lets cover it anyways
        if len(records) > 1:
            # As Cloudflare API cannot filter record containing quotes
            # CAA records must be compared locally
            if self.type == "CAA":
                for rr in records:
                    if (
                        rr["data"]["flags"] == caa_data["flags"]
                        and rr["data"]["tag"] == caa_data["tag"]
                        and rr["data"]["value"] == caa_data["value"]
                    ):
                        return rr, self.changed
            else:
                self.module.fail_json(
                    msg="More than one record already exists for the given attributes. That should be impossible, please open an issue!"
                )
        # record already exists, check if it must be updated
        if len(records) == 1:
            cur_record = records[0]
            do_update = False
            if (self.ttl is not None) and (cur_record["ttl"] != self.ttl):
                do_update = True
            if (self.priority is not None) and ("priority" in cur_record) and (cur_record["priority"] != self.priority):
                do_update = True
            if ("proxied" in new_record) and ("proxied" in cur_record) and (cur_record["proxied"] != self.proxied):
                do_update = True
            if ("data" in new_record) and ("data" in cur_record):
                if cur_record["data"] != new_record["data"]:
                    do_update = True
            if (self.type == "CNAME") and (cur_record["content"] != new_record["content"]):
                do_update = True
            if cur_record["comment"] != new_record["comment"]:
                do_update = True
            if sorted(cur_record["tags"]) != sorted(new_record["tags"]):
                do_update = True
            if do_update:
                if self.module.check_mode:
                    result = new_record
                else:
                    result, info = self._cf_api_call(
                        f"/zones/{zone_id}/dns_records/{records[0]['id']}", "PUT", new_record
                    )
                self.changed = True
                return result, self.changed
            else:
                return records, self.changed
        if self.module.check_mode:
            result = new_record
        else:
            result, info = self._cf_api_call(f"/zones/{zone_id}/dns_records", "POST", new_record)
        self.changed = True
        return result, self.changed


def main():
    module = AnsibleModule(
        argument_spec=dict(
            api_token=dict(type="str", no_log=True, fallback=(env_fallback, ["CLOUDFLARE_TOKEN"])),
            account_api_key=dict(type="str", no_log=True, aliases=["account_api_token"]),
            account_email=dict(type="str"),
            algorithm=dict(type="int"),
            cert_usage=dict(type="int", choices=[0, 1, 2, 3]),
            comment=dict(type="str"),
            hash_type=dict(type="int", choices=[1, 2]),
            key_tag=dict(type="int", no_log=False),
            port=dict(type="int"),
            flag=dict(type="int", choices=[0, 1, 128]),
            tag=dict(type="str", choices=["issue", "issuewild", "iodef"]),
            tags=dict(type="list", elements="str"),
            priority=dict(type="int", default=1),
            proto=dict(type="str"),
            proxied=dict(type="bool", default=False),
            record=dict(type="str", default="@", aliases=["name"]),
            selector=dict(type="int", choices=[0, 1]),
            service=dict(type="str"),
            solo=dict(type="bool"),
            state=dict(type="str", default="present", choices=["absent", "present"]),
            timeout=dict(type="int", default=30),
            ttl=dict(type="int", default=1),
            type=dict(
                type="str",
                choices=["A", "AAAA", "CNAME", "DS", "MX", "NS", "SRV", "SSHFP", "TLSA", "CAA", "TXT", "PTR"],
            ),
            value=dict(type="str", aliases=["content"]),
            weight=dict(type="int", default=1),
            zone=dict(type="str", required=True, aliases=["domain"]),
        ),
        supports_check_mode=True,
        required_if=[
            ("state", "present", ["record", "type", "value"]),
            ("state", "absent", ["record"]),
            ("type", "SRV", ["proto", "service"]),
            ("type", "TLSA", ["proto", "port"]),
            ("type", "CAA", ["flag", "tag"]),
        ],
        required_together=[
            ("account_api_key", "account_email"),
        ],
        required_one_of=[
            ["api_token", "account_api_key"],
        ],
    )

    if module.params["type"] == "SRV":
        if not (
            (
                module.params["weight"] is not None
                and module.params["port"] is not None
                and not (module.params["value"] is None or module.params["value"] == "")
            )
            or (
                module.params["weight"] is None
                and module.params["port"] is None
                and (module.params["value"] is None or module.params["value"] == "")
            )
        ):
            module.fail_json(
                msg="For SRV records the params weight, port and value all need to be defined, or not at all."
            )

    if module.params["type"] == "SSHFP":
        if not (
            (
                module.params["algorithm"] is not None
                and module.params["hash_type"] is not None
                and not (module.params["value"] is None or module.params["value"] == "")
            )
            or (
                module.params["algorithm"] is None
                and module.params["hash_type"] is None
                and (module.params["value"] is None or module.params["value"] == "")
            )
        ):
            module.fail_json(
                msg="For SSHFP records the params algorithm, hash_type and value all need to be defined, or not at all."
            )

    if module.params["type"] == "TLSA":
        if not (
            (
                module.params["cert_usage"] is not None
                and module.params["selector"] is not None
                and module.params["hash_type"] is not None
                and not (module.params["value"] is None or module.params["value"] == "")
            )
            or (
                module.params["cert_usage"] is None
                and module.params["selector"] is None
                and module.params["hash_type"] is None
                and (module.params["value"] is None or module.params["value"] == "")
            )
        ):
            module.fail_json(
                msg="For TLSA records the params cert_usage, selector, hash_type and value all need to be defined, or not at all."
            )

    if module.params["type"] == "CAA":
        if not (
            (
                module.params["flag"] is not None
                and module.params["tag"] is not None
                and not (module.params["value"] is None or module.params["value"] == "")
            )
            or (
                module.params["flag"] is None
                and module.params["tag"] is None
                and (module.params["value"] is None or module.params["value"] == "")
            )
        ):
            module.fail_json(
                msg="For CAA records the params flag, tag and value all need to be defined, or not at all."
            )

    if module.params["type"] == "DS":
        if not (
            (
                module.params["key_tag"] is not None
                and module.params["algorithm"] is not None
                and module.params["hash_type"] is not None
                and not (module.params["value"] is None or module.params["value"] == "")
            )
            or (
                module.params["key_tag"] is None
                and module.params["algorithm"] is None
                and module.params["hash_type"] is None
                and (module.params["value"] is None or module.params["value"] == "")
            )
        ):
            module.fail_json(
                msg="For DS records the params key_tag, algorithm, hash_type and value all need to be defined, or not at all."
            )

    changed = False
    cf_api = CloudflareAPI(module)

    # sanity checks
    if cf_api.is_solo and cf_api.state == "absent":
        module.fail_json(msg="solo=true can only be used with state=present")

    # perform add, delete or update (only the TTL can be updated) of one or
    # more records
    if cf_api.state == "present":
        # delete all records matching record name + type
        if cf_api.is_solo:
            changed = cf_api.delete_dns_records(solo=cf_api.is_solo)
        result, changed = cf_api.ensure_dns_record()
        if isinstance(result, list):
            module.exit_json(changed=changed, result={"record": result[0]})

        module.exit_json(changed=changed, result={"record": result})
    else:
        # force solo to False, just to be sure
        changed = cf_api.delete_dns_records(solo=False)
        module.exit_json(changed=changed)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2020, Shreyas Srish (@shrsr) <ssrish@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_backup_schedule
short_description: Manages backup schedules
description:
- Manage backup schedules on Cisco ACI Multi-Site.
author:
- Akini Ross (@akinross)
options:
  start_date:
    description:
    - The date to start the scheduler in format YYYY-MM-DD
    - If no date is provided, the current date will be used.
    type: str
  start_time:
    description:
    - The time to start the scheduler in format HH:MM:SS
    - If no time is provided, midnight "00:00:00" will be used.
    type: str
  frequency_unit:
    description:
    - The interval unit type
    choices: [ hours, days ]
    type: str
  frequency_length:
    description:
    - Amount of hours or days for the schedule trigger frequency
    type: int
  remote_location:
    description:
    - The remote location's name where the backup should be stored
    type: str
  remote_path:
   description:
    - This path is relative to the remote location.
    - A '/' is automatically added between the remote location folder and this path.
    - This folder structure should already exist on the remote location.
   type: str
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Get current backup schedule
  cisco.mso.mso_backup_schedule:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    state: query
  register: query_result

- name: Set backup schedule
  cisco.mso.mso_backup_schedule:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    frequency_unit: hours
    frequency_length: 7
    remote_location: ansible_test
    state: present

- name: Set backup schedule with date and time
  cisco.mso.mso_backup_schedule:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    frequency_unit: days
    frequency_length: 1
    remote_location: ansible_test
    remote_path: test
    start_time: 20:57:36
    start_date: 2023-04-09
    state: present

- name: Delete backup schedule
  cisco.mso.mso_backup_schedule:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    state: absent
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec
from datetime import datetime, tzinfo, timedelta

# UTC Timezone implementation as datetime.timezone is not supported in Python 2.7


class UTC(tzinfo):
    """UTC"""

    def utcoffset(self, dt):
        return timedelta(0)

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return timedelta(0)


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        start_date=dict(type="str"),
        start_time=dict(type="str"),
        frequency_unit=dict(type="str", choices=["hours", "days"]),
        frequency_length=dict(type="int"),
        remote_location=dict(type="str"),
        remote_path=dict(type="str"),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec, supports_check_mode=True, required_if=[["state", "present", ["frequency_unit", "frequency_length", "remote_location"]]]
    )

    start_date = module.params.get("start_date")
    start_time = module.params.get("start_time")
    frequency_unit = module.params.get("frequency_unit")
    frequency_length = module.params.get("frequency_length")
    remote_location = module.params.get("remote_location")
    remote_path = module.params.get("remote_path")
    state = module.params.get("state")

    mso = MSOModule(module)
    api_path = "backups/schedule"
    mso.existing = mso.request(api_path, method="GET")

    if state == "absent":
        mso.previous = mso.existing
        if module.check_mode:
            mso.existing = {}
        else:
            mso.existing = mso.request(api_path, method="DELETE")

    elif state == "present":
        mso.previous = mso.existing

        remote_location_info = mso.lookup_remote_location(remote_location)

        if start_date:
            try:
                y, m, d = start_date.split("-")
                year = int(y)
                month = int(m)
                day = int(d)
            except Exception as e:
                module.fail_json(msg="Failed to parse date format 'YYYY-MM-DD' %s, %s" % (start_date, e))
        else:
            current_date = datetime.now(UTC()).date()
            year = current_date.year
            month = current_date.month
            day = current_date.day

        if start_time:
            try:
                h, m, s = start_time.split(":")
                hours = int(h)
                minutes = int(m)
                seconds = int(s)
            except Exception as e:
                module.fail_json(msg="Failed to parse time format 'HH:MM:SS' %s, %s" % (start_time, e))
        else:
            hours = minutes = seconds = 0

        try:
            set_date = datetime(year, month, day, hours, minutes, seconds)
        except Exception as e:
            module.fail_json(msg="Failed to create datetime object with date '%s', and time '%s'. Error: %s" % (start_date, start_time, e))

        payload = dict(
            startDate="{0}.000Z".format(set_date.isoformat()),
            intervalTimeUnit=frequency_unit.upper(),
            intervalLength=frequency_length,
            remoteLocationId=remote_location_info.get("id"),
            locationType="remote",
        )

        if remote_path:
            payload.update(remotePath="{0}/{1}".format(remote_location_info.get("path"), remote_path))

        mso.proposed = payload

        if module.check_mode:
            mso.existing = mso.proposed
        else:
            mso.existing = mso.request(api_path, method="POST", data=payload)

    mso.exit_json()


if __name__ == "__main__":
    main()



# This code is part of the Ansible collection community.docker, but is an independent component.
# This particular file, and this file only, is based on the Docker SDK for Python (https://github.com/docker/docker-py/)
#
# Copyright (c) 2016-2025 Docker, Inc.
#
# It is licensed under the Apache 2.0 license (see LICENSES/Apache-2.0.txt in this collection)
# SPDX-License-Identifier: Apache-2.0

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import hashlib
import json
import os

from ..constants import DEFAULT_UNIX_SOCKET, IS_WINDOWS_PLATFORM
from ..utils.config import find_config_file, get_default_config_file
from ..utils.utils import parse_host

METAFILE = "meta.json"


def get_current_context_name_with_source() -> tuple[str, str]:
    if os.environ.get("DOCKER_HOST"):
        return "default", "DOCKER_HOST environment variable set"
    if os.environ.get("DOCKER_CONTEXT"):
        return os.environ["DOCKER_CONTEXT"], "DOCKER_CONTEXT environment variable set"
    docker_cfg_path = find_config_file()
    if docker_cfg_path:
        try:
            with open(docker_cfg_path, "rt", encoding="utf-8") as f:
                return (
                    json.load(f).get("currentContext", "default"),
                    f"configuration file {docker_cfg_path}",
                )
        except Exception:  # pylint: disable=broad-exception-caught
            pass
    return "default", "fallback value"


def get_current_context_name() -> str:
    return get_current_context_name_with_source()[0]


def write_context_name_to_docker_config(name: str | None = None) -> Exception | None:
    if name == "default":
        name = None
    docker_cfg_path = find_config_file()
    config = {}
    if docker_cfg_path:
        try:
            with open(docker_cfg_path, "rt", encoding="utf-8") as f:
                config = json.load(f)
        except Exception as e:  # pylint: disable=broad-exception-caught
            return e
    current_context = config.get("currentContext", None)
    if current_context and not name:
        del config["currentContext"]
    elif name:
        config["currentContext"] = name
    else:
        return None
    if not docker_cfg_path:
        docker_cfg_path = get_default_config_file()
    try:
        with open(docker_cfg_path, "wt", encoding="utf-8") as f:
            json.dump(config, f, indent=4)
        return None
    except Exception as e:  # pylint: disable=broad-exception-caught
        return e


def get_context_id(name: str) -> str:
    return hashlib.sha256(name.encode("utf-8")).hexdigest()


def get_context_dir() -> str:
    docker_cfg_path = find_config_file() or get_default_config_file()
    return os.path.join(os.path.dirname(docker_cfg_path), "contexts")


def get_meta_dir(name: str | None = None) -> str:
    meta_dir = os.path.join(get_context_dir(), "meta")
    if name:
        return os.path.join(meta_dir, get_context_id(name))
    return meta_dir


def get_meta_file(name: str) -> str:
    return os.path.join(get_meta_dir(name), METAFILE)


def get_tls_dir(name: str | None = None, endpoint: str = "") -> str:
    context_dir = get_context_dir()
    if name:
        return os.path.join(context_dir, "tls", get_context_id(name), endpoint)
    return os.path.join(context_dir, "tls")


def get_context_host(path: str | None = None, tls: bool = False) -> str:
    host = parse_host(path, IS_WINDOWS_PLATFORM, tls)
    if host == DEFAULT_UNIX_SOCKET and host.startswith("http+"):
        # remove http+ from default docker socket url
        host = host[5:]
    return host



#!/usr/bin/python

# Copyright 2012 Dag Wieers <dag@wieers.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: hpilo_boot
author: Dag Wieers (@dagwieers)
short_description: Boot system using specific media through HP iLO interface
description:
  - 'This module boots a system through its HP iLO interface. The boot media can be one of: V(cdrom), V(floppy), V(hdd), V(network),
    or V(usb).'
  - This module requires the hpilo python module.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  host:
    description:
      - The HP iLO hostname/address that is linked to the physical system.
    type: str
    required: true
  login:
    description:
      - The login name to authenticate to the HP iLO interface.
    default: Administrator
    type: str
  password:
    description:
      - The password to authenticate to the HP iLO interface.
    default: admin
    type: str
  media:
    description:
      - The boot media to boot the system from.
    choices: ["cdrom", "floppy", "rbsu", "hdd", "network", "normal", "usb"]
    type: str
  image:
    description:
      - The URL of a cdrom, floppy or usb boot media image in the form V(protocol://username:password@hostname:port/filename).
      - V(protocol) is either V(http) or V(https).
      - V(username:password) is optional.
      - V(port) is optional.
    type: str
  state:
    description:
      - The state of the boot media.
      - 'V(no_boot): Do not boot from the device.'
      - 'V(boot_once): Boot from the device once and then notthereafter.'
      - 'V(boot_always): Boot from the device each time the server is rebooted.'
      - 'V(connect): Connect the virtual media device and set to boot_always.'
      - 'V(disconnect): Disconnects the virtual media device and set to no_boot.'
      - 'V(poweroff): Power off the server.'
    default: boot_once
    type: str
    choices: ["boot_always", "boot_once", "connect", "disconnect", "no_boot", "poweroff"]
  force:
    description:
      - Whether to force a reboot (even when the system is already booted).
      - As a safeguard, without force, M(community.general.hpilo_boot) refuses to reboot a server that is already running.
    default: false
    type: bool
  ssl_version:
    description:
      - Change the ssl_version used.
    default: TLSv1
    type: str
    choices: ["SSLv3", "SSLv23", "TLSv1", "TLSv1_1", "TLSv1_2"]
  idempotent_boot_once:
    description:
      - This option makes O(state=boot_once) succeed instead of failing when the server is already powered on.
    type: bool
    default: false
    version_added: 10.6.0
requirements:
  - python-hpilo
notes:
  - To use a USB key image you need to specify floppy as boot media.
  - This module ought to be run from a system that can access the HP iLO interface directly, either by using C(local_action)
    or using C(delegate_to).
"""

EXAMPLES = r"""
- name: Task to boot a system using an ISO from an HP iLO interface only if the system is an HP server
  community.general.hpilo_boot:
    host: YOUR_ILO_ADDRESS
    login: YOUR_ILO_LOGIN
    password: YOUR_ILO_PASSWORD
    media: cdrom
    image: http://some-web-server/iso/boot.iso
  when: cmdb_hwmodel.startswith('HP ')
  delegate_to: localhost

- name: Power off a server
  community.general.hpilo_boot:
    host: YOUR_ILO_HOST
    login: YOUR_ILO_LOGIN
    password: YOUR_ILO_PASSWORD
    state: poweroff
  delegate_to: localhost
"""

RETURN = r"""
# Default return values
"""

import time
import traceback
import warnings

HPILO_IMP_ERR = None
try:
    import hpilo

    HAS_HPILO = True
except ImportError:
    HPILO_IMP_ERR = traceback.format_exc()
    HAS_HPILO = False

from ansible.module_utils.basic import AnsibleModule, missing_required_lib

# Suppress warnings from hpilo
warnings.simplefilter("ignore")


def main():
    module = AnsibleModule(
        argument_spec=dict(
            host=dict(type="str", required=True),
            login=dict(type="str", default="Administrator"),
            password=dict(type="str", default="admin", no_log=True),
            media=dict(type="str", choices=["cdrom", "floppy", "rbsu", "hdd", "network", "normal", "usb"]),
            image=dict(type="str"),
            state=dict(
                type="str",
                default="boot_once",
                choices=["boot_always", "boot_once", "connect", "disconnect", "no_boot", "poweroff"],
            ),
            force=dict(type="bool", default=False),
            idempotent_boot_once=dict(type="bool", default=False),
            ssl_version=dict(type="str", default="TLSv1", choices=["SSLv3", "SSLv23", "TLSv1", "TLSv1_1", "TLSv1_2"]),
        )
    )

    if not HAS_HPILO:
        module.fail_json(msg=missing_required_lib("python-hpilo"), exception=HPILO_IMP_ERR)

    host = module.params["host"]
    login = module.params["login"]
    password = module.params["password"]
    media = module.params["media"]
    image = module.params["image"]
    state = module.params["state"]
    force = module.params["force"]
    idempotent_boot_once = module.params["idempotent_boot_once"]
    ssl_version = getattr(hpilo.ssl, f"PROTOCOL_{module.params.get('ssl_version').upper().replace('V', 'v')}")

    ilo = hpilo.Ilo(host, login=login, password=password, ssl_version=ssl_version)
    changed = False
    status = {}
    power_status = "UNKNOWN"

    if media and state in ("boot_always", "boot_once", "connect", "disconnect", "no_boot"):
        # Workaround for: Error communicating with iLO: Problem manipulating EV
        try:
            ilo.set_one_time_boot(media)
        except hpilo.IloError:
            time.sleep(60)
            ilo.set_one_time_boot(media)

        # TODO: Verify if image URL exists/works
        if image:
            ilo.insert_virtual_media(media, image)
            changed = True

        if media == "cdrom":
            ilo.set_vm_status("cdrom", state, True)
            status = ilo.get_vm_status()
            changed = True
        elif media in ("floppy", "usb"):
            ilo.set_vf_status(state, True)
            status = ilo.get_vf_status()
            changed = True

    # Only perform a boot when state is boot_once or boot_always, or in case we want to force a reboot
    if state in ("boot_once", "boot_always") or force:
        power_status = ilo.get_host_power_status()

        if power_status == "ON":
            if not force and not idempotent_boot_once:
                # module.deprecate(
                #     'The failure of the module when the server is already powered on is being deprecated.'
                #     ' Please set the parameter "idempotent_boot_once=true" to start using the new behavior.',
                #     version='11.0.0',
                #     collection_name='community.general'
                # )
                module.fail_json(msg=f"HP iLO ({host}) reports that the server is already powered on !")
            elif not force and idempotent_boot_once:
                pass
            elif force:
                ilo.warm_boot_server()
                #            ilo.cold_boot_server()
                changed = True
        else:
            ilo.press_pwr_btn()
            #            ilo.reset_server()
            #            ilo.set_host_power(host_power=True)
            changed = True

    elif state in ("poweroff"):
        power_status = ilo.get_host_power_status()

        if power_status != "OFF":
            ilo.hold_pwr_btn()
            #            ilo.set_host_power(host_power=False)
            changed = True

    module.exit_json(changed=changed, power=power_status, **status)


if __name__ == "__main__":
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2023 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The nxos_fc_interfaces config file.
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to its desired end-state is
created.
"""


from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module import (
    ResourceModule,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    dict_merge,
)

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.facts.facts import Facts
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.rm_templates.fc_interfaces import (
    Fc_interfacesTemplate,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.utils.utils import (
    normalize_interface,
)


class Fc_interfaces(ResourceModule):
    """
    The nxos_fc_interfaces config class
    """

    def __init__(self, module):
        super(Fc_interfaces, self).__init__(
            empty_fact_val={},
            facts_module=Facts(module, chassis_type="mds"),
            module=module,
            resource="fc_interfaces",
            tmplt=Fc_interfacesTemplate(),
        )
        self.parsers = ["description", "speed", "mode", "trunk_mode", "analytics"]

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        if self.state not in ["parsed", "gathered"]:
            self.generate_commands()
            self.run_commands()
        return self.result

    def generate_commands(self):
        """Generate configuration commands to send based on
        want, have and desired state.
        """

        wantd = {entry["name"]: entry for entry in self.want}
        haved = {entry["name"]: entry for entry in self.have}

        for each in wantd, haved:
            self.normalize_interface_names(each)

        # if state is merged, merge want onto have and then compare
        if self.state == "merged":
            wantd = dict_merge(haved, wantd)

        # if state is deleted, empty out wantd and set haved to wantd
        if self.state == "deleted":
            haved = {k: v for k, v in haved.items() if k in wantd or not wantd}
            wantd = {}

        # remove superfluous config for overridden and deleted
        if self.state in ["overridden", "deleted"]:
            for k, have in haved.items():
                if k not in wantd:
                    self._compare(want={}, have=have)

        for k, want in wantd.items():
            self._compare(want=want, have=haved.pop(k, {}))

        modified_list = [
            "switchport trunk mode on" if item.startswith("no switchport trunk mode") else item
            for item in self.commands
        ]
        self.commands = modified_list

    def _calculate_ana_config(self, want_ana, have_ana):
        """
        get the cmds based on want_ana and have_ana and the state

        Args:
            want_ana (str): analytics type which you want
            have_ana (str): analytics type which you have

        +----------+----------+---------+
        |            MERGED             |
        |----------+----------+---------+
        | want_ana | have_ana | outcome |
        +----------+----------+---------+
        | ""       | *        | no op   |
        | fc-scsi  | *        | fc-scsi |
        | fc-scsi  | fc-all   | no op   |
        | fc-nvme  | *        | fc-nvme |
        | fc-nvme  | fc-all   | no op   |
        +----------+----------+---------+


        +----------+----------+-----------+
        |            DELETED              |
        |----------+----------+-----------+
        | want_ana | have_ana | outcome   |
        +----------+----------+-----------+
        | *        | fc-scsi  | no fc-all |
        | *        | fc-nvme  | no fc-all |
        | *        | fc-all   | no fc-all |
        | *        | ""       | no op     |
        +----------+----------+-----------+


        +----------+----------+---------------------+
        |            REPLACED/OVERRIDEN             |
        |----------+----------+---------------------+
        | want_ana | have_ana | outcome             |
        +----------+----------+---------------------+
        | ""       | *        | no fc-all           |
        | fc-scsi  | ""       | fc-scsi             |
        | fc-nvme  | ""       | fc-nvme             |
        | fc-all   | ""       | fc-all              |
        | fc-scsi  | *        | no fc-all ; fc-scsi |
        | fc-nvme  | *        | no fc-all ; fc-nvme |
        | fc-all   | *        | fc-all              |
        +----------+----------+---------------------+


        """

        if want_ana == have_ana:
            return []
        val = []
        if self.state in ["overridden", "replaced"]:
            if want_ana == "":
                val = ["no analytics type fc-all"]
            elif want_ana == "fc-all":
                val = ["analytics type fc-all"]
            elif have_ana == "":
                val = [f"analytics type {want_ana}"]
            else:
                val = ["no analytics type fc-all", f"analytics type {want_ana}"]
        elif self.state in ["deleted"]:
            if have_ana:
                val = ["no analytics type fc-all"]
        elif self.state in ["merged"]:
            if want_ana:
                if have_ana != "fc-all":
                    val = [f"analytics type {want_ana}"]
        return val

    def _compare(self, want, have):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the Fc_interfaces network resource.
        """

        begin = len(self.commands)
        self.compare(parsers=self.parsers, want=want, have=have)
        if want.get("enabled") != have.get("enabled"):
            if want.get("enabled"):
                self.addcmd(want, "enabled", True)
            else:
                if want:
                    self.addcmd(want, "enabled", False)
                elif have.get("enabled"):
                    # handles deleted as want be blank and only
                    # negates if no shutdown
                    self.addcmd(have, "enabled", False)

        ana_cmds = self._calculate_ana_config(want.get("analytics", ""), have.get("analytics", ""))

        self.commands.extend(ana_cmds)

        if len(self.commands) != begin:
            self.commands.insert(begin, self._tmplt.render(want or have, "interface", False))

    def normalize_interface_names(self, param):
        if param:
            for _k, val in param.items():
                val["name"] = normalize_interface(val["name"])
        return param



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright: (c) 2025, Ciril Troxler <ciril.troxler@cloudscale.ch>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: volume_snapshot
short_description: Manage volume snapshots on the cloudscale.ch IaaS service
description:
  - Get, create, update, delete volume snapshots on the cloudscale.ch IaaS service.
notes:
  - If I(uuid) option is provided, it takes precedence over I(name) for volume snapshot selection.
    This allows to update the volume snapshot's name.
  - If no I(uuid) option is provided, I(name) is used for volume snapshot selection.
    If more than one volume snapshot with this name exists, execution is aborted.
  - To revert a snapshot use the volume module.
author:
  - Ciril Troxler (@ctx)
version_added: "2.5.0"
options:
  state:
    description:
      - State of the volume snapshot.
    choices: [ present, absent ]
    default: present
    type: str
  name:
    description:
      - Name of the volume snapshot. Either I(name) or I(uuid) are required.
    type: str
  uuid:
    description:
      - UUID of the volume snapshot.
    type: str
  source_volume:
    description:
      - UUID of the volume this snapshot belongs to.
        If I(name) and I(source_volume) are present, a new volume snapshot is created.
        This parameter has no effect on existing volume snapshots (I(uuid) option is present).
    type: str
  tags:
    description:
      - Tags assigned to the volume snapshot. Set this to C({}) to clear any tags.
    type: dict
extends_documentation_fragment: cloudscale_ch.cloud.api_parameters
'''

EXAMPLES = '''
# Create a volume snapshot for a volume
- name: Create a volume snapshot
  cloudscale_ch.cloud.volume_snapshot:
    name: 'pre-dist-upgrade'
    source_volume: '2db69ba3-1864-4608-853a-0771b6885a3a'
    tags: {}
    api_token: xxxxx

# Get a volume snapshot by name
- name: Get facts of a volume snapshot
  cloudscale_ch.cloud.volume_snapshot:
    name: 'pre-dist-upgrade'
    api_token: xxxxx

# Delete a volume snapshot
- name: Delete a volume snapshot
  cloudscale_ch.cloud.volume_snapshot:
    uuid: '351d461c-2333-455f-b788-db11bf0b4aa2'
    state: absent
'''

RETURN = '''
href:
  description: The API URL to get details about this volume snapshot.
  returned: state == present
  type: str
  sample: https://api.cloudscale.ch/v1/volume-snapshots/e504dc99-ff01-4e89-ad89-7df080f97b4b
uuid:
  description: The unique identifier for this volume snapshot.
  returned: state == present
  type: str
  sample: e504dc99-ff01-4e89-ad89-7df080f97b4b
name:
  description: The display name of the volume snapshot.
  returned: state == present
  type: str
  sample: my_ssd_volume_snapshot
created_at:
  description: The creation date and time of the resource.
  returned: state == present
  type: str
  sample: "2025-04-10T11:05:45.777073Z"
size_gb:
  description: The size of the volume in GB.
  returned: state == present
  type: str
  sample: 50
source_volume:
  description: The source volume this volume snapshot belongs to.
  returned: state == present
  type: dict
  sample:
    "source_volume": {
        "href": "https://api.cloudscale.ch/v1/volumes/0952b1ca-3657-4901-b4c8-669796ec55e6",
        "name": "my_ssd_volume",
        "uuid": "0952b1ca-3657-4901-b4c8-669796ec55e6"
    }
state:
  description: The current status of the volume.
  returned: success
  type: str
  sample: present
status:
  description: The current status of the volume snapshot.
  returned: success
  type: str
  sample: available
tags:
  description: Tags associated with the volume.
  returned: state == present
  type: dict
  sample: { 'project': 'my project' }
'''

from ansible.module_utils.basic import AnsibleModule
from ..module_utils.api import (
    AnsibleCloudscaleBase,
    cloudscale_argument_spec,
)


class AnsibleCloudscaleVolumeSnapshot(AnsibleCloudscaleBase):

    def __init__(self, module):
        super(AnsibleCloudscaleVolumeSnapshot, self).__init__(
            module,
            resource_name='volume-snapshots',
            resource_create_param_keys=[
                'name',
                'source_volume',
                'tags',
            ],
            resource_update_param_keys=[
                'uuid',
                'tags',
            ],
        )

    def absent(self):
        resource = super().absent()
        if not self._module.check_mode:
            self.wait_for_state('state', 'absent')
        return resource


def main():
    argument_spec = cloudscale_argument_spec()
    argument_spec.update(dict(
        name=dict(type='str'),
        uuid=dict(type='str'),
        source_volume=dict(type='str'),
        tags=dict(type='dict'),
        state=dict(type='str', default='present', choices=('present', 'absent')),
    ))

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=(),
        required_one_of=(('name', 'uuid'),),
        required_if=(('state', 'absent', ('uuid',),),),
        supports_check_mode=True,
    )

    cloudscale_volume_snapshot = AnsibleCloudscaleVolumeSnapshot(module)

    if module.params['state'] == 'absent':
        result = cloudscale_volume_snapshot.absent()
    else:
        result = cloudscale_volume_snapshot.present()
    module.exit_json(**result)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright 2024 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

#############################################
#                WARNING                    #
#############################################
#
# This file is auto generated by the
# ansible.content_builder.
#
# Manually editing this file is not advised.
#
# To update the argspec make the desired changes
# in the documentation in the module file and re-run
# ansible.content_builder commenting out
# the path to external 'docstring' in build.yaml.
#
##############################################

"""
The arg spec for the iosxr_route_maps module
"""


class Route_mapsArgs(object):  # pylint: disable=R0903
    """The arg spec for the iosxr_route_maps module"""

    argument_spec = {
        "config": {
            "type": "list",
            "elements": "dict",
            "options": {
                "name": {"type": "str"},
                "global": {
                    "type": "dict",
                    "options": {
                        "add": {
                            "type": "dict",
                            "options": {
                                "eigrp_metric": {
                                    "type": "dict",
                                    "options": {
                                        "bandwidth": {"type": "int"},
                                        "delay": {"type": "int"},
                                        "reliability": {"type": "int"},
                                        "effective_bandwith": {"type": "int"},
                                        "max_transmission": {"type": "int"},
                                    },
                                },
                                "rip_metric": {"type": "int"},
                            },
                        },
                        "apply": {
                            "type": "list",
                            "elements": "dict",
                            "options": {
                                "route_policy": {"type": "str"},
                                "route_policy_input": {"type": "str"},
                            },
                        },
                        "drop": {"type": "bool"},
                        "pass": {"type": "bool"},
                        "prepend": {
                            "type": "dict",
                            "options": {
                                "number_of_times": {"type": "int"},
                                "as_path": {"type": "int"},
                                "most_recent": {"type": "bool"},
                                "own_as": {"type": "bool"},
                            },
                        },
                        "suppress_route": {"type": "bool"},
                        "unsuppress_route": {"type": "bool"},
                        "remove": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "entire_aspath": {"type": "bool"},
                            },
                        },
                        "set": {
                            "type": "dict",
                            "options": {
                                "administrative_distance": {"type": "int"},
                                "aigp_metric": {
                                    "type": "dict",
                                    "options": {
                                        "icrement": {"type": "bool"},
                                        "decrement": {"type": "bool"},
                                        "metric_number": {"type": "int"},
                                        "igp_cost": {"type": "bool"},
                                    },
                                },
                                "local_preference": {
                                    "type": "list",
                                    "elements": "dict",
                                    "options": {
                                        "multiply": {"type": "bool"},
                                        "increment": {"type": "bool"},
                                        "decrement": {"type": "bool"},
                                        "metric_number": {
                                            "type": "int",
                                            "required": True,
                                        },
                                    },
                                },
                                "attribute_set": {"type": "str"},
                                "c_multicast_routing": {
                                    "type": "dict",
                                    "options": {
                                        "bgp": {"type": "bool"},
                                        "pim": {"type": "bool"},
                                    },
                                },
                                "community": {
                                    "type": "dict",
                                    "options": {
                                        "community_name": {"type": "str"},
                                        "additive": {"type": "bool"},
                                    },
                                },
                                "core_tree": {
                                    "type": "dict",
                                    "options": {
                                        "ingress_replication": {"type": "bool"},
                                        "ingress_replication_default": {
                                            "type": "bool",
                                        },
                                        "ingress_replication_partitioned": {
                                            "type": "bool",
                                        },
                                        "mldp": {"type": "bool"},
                                        "mldp_default": {"type": "bool"},
                                        "mldp_inband": {"type": "bool"},
                                        "mldp_partitioned_mp2mp": {"type": "bool"},
                                        "mldp_partitioned_p2mp": {"type": "bool"},
                                        "p2mp_te": {"type": "bool"},
                                        "p2mp_te_default": {"type": "bool"},
                                        "p2mp_te_partitioned": {"type": "bool"},
                                        "pim_default": {"type": "bool"},
                                        "sr_p2mp": {"type": "bool"},
                                    },
                                },
                                "dampening": {
                                    "type": "dict",
                                    "options": {
                                        "halflife": {"type": "int"},
                                        "max_suppress": {"type": "int"},
                                        "reuse": {"type": "int"},
                                        "suppress": {"type": "int"},
                                    },
                                },
                                "downstream_core_tree": {
                                    "type": "dict",
                                    "options": {
                                        "ingress_replication": {"type": "bool"},
                                        "mldp": {"type": "bool"},
                                        "p2mp_te": {"type": "bool"},
                                        "sr_p2mp": {"type": "bool"},
                                    },
                                },
                                "eigrp_metric": {
                                    "type": "dict",
                                    "options": {
                                        "bandwidth": {"type": "int"},
                                        "delay": {"type": "int"},
                                        "reliability": {"type": "int"},
                                        "effective_bandwith": {"type": "int"},
                                        "max_transmission": {"type": "int"},
                                    },
                                },
                                "extcommunity": {
                                    "type": "dict",
                                    "options": {
                                        "soo": {"type": "str"},
                                        "rt": {"type": "str"},
                                        "additive": {"type": "bool"},
                                        "bandwidth": {"type": "str"},
                                        "color": {"type": "str"},
                                        "cost": {"type": "str"},
                                        "redirect_to_rt": {"type": "str"},
                                        "seg_nh": {"type": "str"},
                                    },
                                },
                                "fallback_vrf_lookup": {"type": "bool"},
                                "flow_tag": {"type": "int"},
                                "forward_class": {"type": "int"},
                                "ip_precedence": {"type": "int"},
                                "isis_metric": {"type": "int"},
                                "label": {"type": "int"},
                                "label_index": {"type": "int"},
                                "label_mode": {
                                    "type": "dict",
                                    "options": {
                                        "per_ce": {"type": "bool"},
                                        "per_prefix": {"type": "bool"},
                                        "per_vrf": {"type": "bool"},
                                    },
                                },
                                "large_community": {"type": "str"},
                                "level": {
                                    "type": "dict",
                                    "options": {
                                        "level_1": {"type": "bool"},
                                        "level_1_2": {"type": "bool"},
                                        "level_2": {"type": "bool"},
                                    },
                                },
                                "load_balance": {"type": "bool"},
                                "lsm_root": {"type": "str"},
                                "med": {
                                    "type": "dict",
                                    "options": {
                                        "value": {"type": "int"},
                                        "increment": {"type": "int"},
                                        "decrement": {"type": "int"},
                                        "igp_cost": {"type": "bool"},
                                        "max_reachable": {"type": "bool"},
                                        "parameter": {"type": "str"},  # like "$param1"
                                    },
                                },
                                "metric_type": {
                                    "type": "dict",
                                    "options": {
                                        "external": {"type": "bool"},
                                        "internal": {"type": "bool"},
                                        "rib_metric_as_external": {"type": "bool"},
                                        "rib_metric_as_internal": {"type": "bool"},
                                        "type_1": {"type": "bool"},
                                        "type_2": {"type": "bool"},
                                    },
                                },
                                "mpls": {"type": "str"},
                                "next_hop": {
                                    "type": "dict",
                                    "options": {"address": {"type": "str"}},
                                },
                                "origin": {
                                    "type": "dict",
                                    "options": {
                                        "egp": {"type": "bool"},
                                        "igp": {"type": "bool"},
                                        "rincomplete": {"type": "bool"},
                                    },
                                },
                                "ospf_metric": {"type": "int"},
                                "path_selection": {
                                    "type": "dict",
                                    "options": {
                                        "all": {"type": "bool"},
                                        "backup": {
                                            "type": "dict",
                                            "options": {
                                                "backup_decimal": {"type": "int"},
                                                "advertise": {"type": "bool"},
                                                "install": {"type": "bool"},
                                            },
                                        },
                                        "best_path": {"type": "bool"},
                                        "group_best": {"type": "bool"},
                                        "multiplath": {"type": "bool"},
                                    },
                                },
                                "path_color": {"type": "bool"},
                                "qos_group": {"type": "int"},
                                "rib_metric": {"type": "int"},
                                "rip_metric": {"type": "int"},
                                "rip_tag": {"type": "int"},
                                "rt_set": {"type": "int"},
                                "s_pmsi": {"type": "bool"},
                                "spf_priority": {
                                    "type": "dict",
                                    "options": {
                                        "critical": {"type": "bool"},
                                        "high": {"type": "bool"},
                                        "medium": {"type": "bool"},
                                    },
                                },
                                "static_p2mp_te": {"type": "str"},
                                "tag": {"type": "int"},
                                "traffic_index": {
                                    "type": "dict",
                                    "options": {
                                        "index_number": {"type": "int"},
                                        "ignore": {"type": "bool"},
                                    },
                                },
                                "upstream_core_tree": {
                                    "type": "dict",
                                    "options": {
                                        "ingress_replication": {"type": "bool"},
                                        "mldp": {"type": "bool"},
                                        "p2mp_te": {"type": "bool"},
                                        "sr_p2mp": {"type": "bool"},
                                    },
                                },
                                "vpn_distinguisher": {"type": "int"},
                                "weight": {"type": "int"},
                            },
                        },
                    },
                },
                "if_section": {
                    "type": "dict",
                    "options": {
                        "condition": {"type": "str"},
                        "add": {
                            "type": "dict",
                            "options": {
                                "eigrp_metric": {
                                    "type": "dict",
                                    "options": {
                                        "bandwidth": {"type": "int"},
                                        "delay": {"type": "int"},
                                        "reliability": {"type": "int"},
                                        "effective_bandwith": {"type": "int"},
                                        "max_transmission": {"type": "int"},
                                    },
                                },
                                "rip_metric": {"type": "int"},
                            },
                        },
                        "apply": {
                            "type": "list",
                            "elements": "dict",
                            "options": {
                                "route_policy": {"type": "str"},
                                "route_policy_input": {"type": "str"},
                            },
                        },
                        "drop": {"type": "bool"},
                        "pass": {"type": "bool"},
                        "prepend": {
                            "type": "dict",
                            "options": {
                                "number_of_times": {"type": "int"},
                                "as_path": {"type": "int"},
                                "most_recent": {"type": "bool"},
                                "own_as": {"type": "bool"},
                            },
                        },
                        "suppress_route": {"type": "bool"},
                        "unsuppress_route": {"type": "bool"},
                        "remove": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "entire_aspath": {"type": "bool"},
                            },
                        },
                        "set": {
                            "type": "dict",
                            "options": {
                                "administrative_distance": {"type": "int"},
                                "local_preference": {
                                    "type": "list",
                                    "elements": "dict",
                                    "options": {
                                        "multiply": {"type": "bool"},
                                        "increment": {"type": "bool"},
                                        "decrement": {"type": "bool"},
                                        "metric_number": {
                                            "type": "int",
                                            "required": True,
                                        },
                                    },
                                },
                                "aigp_metric": {
                                    "type": "dict",
                                    "options": {
                                        "icrement": {"type": "bool"},
                                        "decrement": {"type": "bool"},
                                        "metric_number": {"type": "int"},
                                        "igp_cost": {"type": "bool"},
                                    },
                                },
                                "attribute_set": {"type": "str"},
                                "c_multicast_routing": {
                                    "type": "dict",
                                    "options": {
                                        "bgp": {"type": "bool"},
                                        "pim": {"type": "bool"},
                                    },
                                },
                                "community": {
                                    "type": "dict",
                                    "options": {
                                        "community_name": {"type": "str"},
                                        "additive": {"type": "bool"},
                                    },
                                },
                                "core_tree": {
                                    "type": "dict",
                                    "options": {
                                        "ingress_replication": {"type": "bool"},
                                        "ingress_replication_default": {
                                            "type": "bool",
                                        },
                                        "ingress_replication_partitioned": {
                                            "type": "bool",
                                        },
                                        "mldp": {"type": "bool"},
                                        "mldp_default": {"type": "bool"},
                                        "mldp_inband": {"type": "bool"},
                                        "mldp_partitioned_mp2mp": {"type": "bool"},
                                        "mldp_partitioned_p2mp": {"type": "bool"},
                                        "p2mp_te": {"type": "bool"},
                                        "p2mp_te_default": {"type": "bool"},
                                        "p2mp_te_partitioned": {"type": "bool"},
                                        "pim_default": {"type": "bool"},
                                        "sr_p2mp": {"type": "bool"},
                                    },
                                },
                                "dampening": {
                                    "type": "dict",
                                    "options": {
                                        "halflife": {"type": "int"},
                                        "max_suppress": {"type": "int"},
                                        "reuse": {"type": "int"},
                                        "suppress": {"type": "int"},
                                    },
                                },
                                "downstream_core_tree": {
                                    "type": "dict",
                                    "options": {
                                        "ingress_replication": {"type": "bool"},
                                        "mldp": {"type": "bool"},
                                        "p2mp_te": {"type": "bool"},
                                        "sr_p2mp": {"type": "bool"},
                                    },
                                },
                                "eigrp_metric": {
                                    "type": "dict",
                                    "options": {
                                        "bandwidth": {"type": "int"},
                                        "delay": {"type": "int"},
                                        "reliability": {"type": "int"},
                                        "effective_bandwith": {"type": "int"},
                                        "max_transmission": {"type": "int"},
                                    },
                                },
                                "extcommunity": {
                                    "type": "dict",
                                    "options": {
                                        "soo": {"type": "str"},
                                        "rt": {"type": "str"},
                                        "additive": {"type": "bool"},
                                        "bandwidth": {"type": "str"},
                                        "color": {"type": "str"},
                                        "cost": {"type": "str"},
                                        "redirect_to_rt": {"type": "str"},
                                        "seg_nh": {"type": "str"},
                                    },
                                },
                                "fallback_vrf_lookup": {"type": "bool"},
                                "flow_tag": {"type": "int"},
                                "forward_class": {"type": "int"},
                                "ip_precedence": {"type": "int"},
                                "isis_metric": {"type": "int"},
                                "label": {"type": "int"},
                                "label_index": {"type": "int"},
                                "label_mode": {
                                    "type": "dict",
                                    "options": {
                                        "per_ce": {"type": "bool"},
                                        "per_prefix": {"type": "bool"},
                                        "per_vrf": {"type": "bool"},
                                    },
                                },
                                "large_community": {"type": "str"},
                                "level": {
                                    "type": "dict",
                                    "options": {
                                        "level_1": {"type": "bool"},
                                        "level_1_2": {"type": "bool"},
                                        "level_2": {"type": "bool"},
                                    },
                                },
                                "load_balance": {"type": "bool"},
                                "lsm_root": {"type": "str"},
                                "metric_type": {
                                    "type": "dict",
                                    "options": {
                                        "external": {"type": "bool"},
                                        "internal": {"type": "bool"},
                                        "rib_metric_as_external": {"type": "bool"},
                                        "rib_metric_as_internal": {"type": "bool"},
                                        "type_1": {"type": "bool"},
                                        "type_2": {"type": "bool"},
                                    },
                                },
                                "med": {
                                    "type": "dict",
                                    "options": {
                                        "value": {"type": "int"},
                                        "increment": {"type": "int"},
                                        "decrement": {"type": "int"},
                                        "igp_cost": {"type": "bool"},
                                        "max_reachable": {"type": "bool"},
                                        "parameter": {"type": "str"},  # like "$param1"
                                    },
                                },
                                "mpls": {"type": "str"},
                                "next_hop": {
                                    "type": "dict",
                                    "options": {"address": {"type": "str"}},
                                },
                                "origin": {
                                    "type": "dict",
                                    "options": {
                                        "egp": {"type": "bool"},
                                        "igp": {"type": "bool"},
                                        "rincomplete": {"type": "bool"},
                                    },
                                },
                                "ospf_metric": {"type": "int"},
                                "path_selection": {
                                    "type": "dict",
                                    "options": {
                                        "all": {"type": "bool"},
                                        "backup": {
                                            "type": "dict",
                                            "options": {
                                                "backup_decimal": {"type": "int"},
                                                "advertise": {"type": "bool"},
                                                "install": {"type": "bool"},
                                            },
                                        },
                                        "best_path": {"type": "bool"},
                                        "group_best": {"type": "bool"},
                                        "multiplath": {"type": "bool"},
                                    },
                                },
                                "path_color": {"type": "bool"},
                                "qos_group": {"type": "int"},
                                "rib_metric": {"type": "int"},
                                "rip_metric": {"type": "int"},
                                "rip_tag": {"type": "int"},
                                "rt_set": {"type": "int"},
                                "s_pmsi": {"type": "bool"},
                                "spf_priority": {
                                    "type": "dict",
                                    "options": {
                                        "critical": {"type": "bool"},
                                        "high": {"type": "bool"},
                                        "medium": {"type": "bool"},
                                    },
                                },
                                "static_p2mp_te": {"type": "str"},
                                "tag": {"type": "int"},
                                "traffic_index": {
                                    "type": "dict",
                                    "options": {
                                        "index_number": {"type": "int"},
                                        "ignore": {"type": "bool"},
                                    },
                                },
                                "upstream_core_tree": {
                                    "type": "dict",
                                    "options": {
                                        "ingress_replication": {"type": "bool"},
                                        "mldp": {"type": "bool"},
                                        "p2mp_te": {"type": "bool"},
                                        "sr_p2mp": {"type": "bool"},
                                    },
                                },
                                "vpn_distinguisher": {"type": "int"},
                                "weight": {"type": "int"},
                            },
                        },
                    },
                },
                "elseif_section": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "condition": {"type": "str"},
                        "add": {
                            "type": "dict",
                            "options": {
                                "eigrp_metric": {
                                    "type": "dict",
                                    "options": {
                                        "bandwidth": {"type": "int"},
                                        "delay": {"type": "int"},
                                        "reliability": {"type": "int"},
                                        "effective_bandwith": {"type": "int"},
                                        "max_transmission": {"type": "int"},
                                    },
                                },
                                "rip_metric": {"type": "int"},
                            },
                        },
                        "apply": {
                            "type": "list",
                            "elements": "dict",
                            "options": {
                                "route_policy": {"type": "str"},
                                "route_policy_input": {"type": "str"},
                            },
                        },
                        "drop": {"type": "bool"},
                        "pass": {"type": "bool"},
                        "prepend": {
                            "type": "dict",
                            "options": {
                                "number_of_times": {"type": "int"},
                                "as_path": {"type": "int"},
                                "most_recent": {"type": "bool"},
                                "own_as": {"type": "bool"},
                            },
                        },
                        "suppress_route": {"type": "bool"},
                        "unsuppress_route": {"type": "bool"},
                        "remove": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "entire_aspath": {"type": "bool"},
                            },
                        },
                        "set": {
                            "type": "dict",
                            "options": {
                                "administrative_distance": {"type": "int"},
                                "local_preference": {
                                    "type": "list",
                                    "elements": "dict",
                                    "options": {
                                        "multiply": {"type": "bool"},
                                        "increment": {"type": "bool"},
                                        "decrement": {"type": "bool"},
                                        "metric_number": {
                                            "type": "int",
                                            "required": True,
                                        },
                                    },
                                },
                                "aigp_metric": {
                                    "type": "dict",
                                    "options": {
                                        "icrement": {"type": "bool"},
                                        "decrement": {"type": "bool"},
                                        "metric_number": {"type": "int"},
                                        "igp_cost": {"type": "bool"},
                                    },
                                },
                                "attribute_set": {"type": "str"},
                                "c_multicast_routing": {
                                    "type": "dict",
                                    "options": {
                                        "bgp": {"type": "bool"},
                                        "pim": {"type": "bool"},
                                    },
                                },
                                "community": {
                                    "type": "dict",
                                    "options": {
                                        "community_name": {"type": "str"},
                                        "additive": {"type": "bool"},
                                    },
                                },
                                "core_tree": {
                                    "type": "dict",
                                    "options": {
                                        "ingress_replication": {"type": "bool"},
                                        "ingress_replication_default": {
                                            "type": "bool",
                                        },
                                        "ingress_replication_partitioned": {
                                            "type": "bool",
                                        },
                                        "mldp": {"type": "bool"},
                                        "mldp_default": {"type": "bool"},
                                        "mldp_inband": {"type": "bool"},
                                        "mldp_partitioned_mp2mp": {"type": "bool"},
                                        "mldp_partitioned_p2mp": {"type": "bool"},
                                        "p2mp_te": {"type": "bool"},
                                        "p2mp_te_default": {"type": "bool"},
                                        "p2mp_te_partitioned": {"type": "bool"},
                                        "pim_default": {"type": "bool"},
                                        "sr_p2mp": {"type": "bool"},
                                    },
                                },
                                "dampening": {
                                    "type": "dict",
                                    "options": {
                                        "halflife": {"type": "int"},
                                        "max_suppress": {"type": "int"},
                                        "reuse": {"type": "int"},
                                        "suppress": {"type": "int"},
                                    },
                                },
                                "downstream_core_tree": {
                                    "type": "dict",
                                    "options": {
                                        "ingress_replication": {"type": "bool"},
                                        "mldp": {"type": "bool"},
                                        "p2mp_te": {"type": "bool"},
                                        "sr_p2mp": {"type": "bool"},
                                    },
                                },
                                "eigrp_metric": {
                                    "type": "dict",
                                    "options": {
                                        "bandwidth": {"type": "int"},
                                        "delay": {"type": "int"},
                                        "reliability": {"type": "int"},
                                        "effective_bandwith": {"type": "int"},
                                        "max_transmission": {"type": "int"},
                                    },
                                },
                                "extcommunity": {
                                    "type": "dict",
                                    "options": {
                                        "soo": {"type": "str"},
                                        "rt": {"type": "str"},
                                        "additive": {"type": "bool"},
                                        "bandwidth": {"type": "str"},
                                        "color": {"type": "str"},
                                        "cost": {"type": "str"},
                                        "redirect_to_rt": {"type": "str"},
                                        "seg_nh": {"type": "str"},
                                    },
                                },
                                "fallback_vrf_lookup": {"type": "bool"},
                                "flow_tag": {"type": "int"},
                                "forward_class": {"type": "int"},
                                "ip_precedence": {"type": "int"},
                                "isis_metric": {"type": "int"},
                                "label": {"type": "int"},
                                "label_index": {"type": "int"},
                                "label_mode": {
                                    "type": "dict",
                                    "options": {
                                        "per_ce": {"type": "bool"},
                                        "per_prefix": {"type": "bool"},
                                        "per_vrf": {"type": "bool"},
                                    },
                                },
                                "large_community": {"type": "str"},
                                "level": {
                                    "type": "dict",
                                    "options": {
                                        "level_1": {"type": "bool"},
                                        "level_1_2": {"type": "bool"},
                                        "level_2": {"type": "bool"},
                                    },
                                },
                                "load_balance": {"type": "bool"},
                                "lsm_root": {"type": "str"},
                                "metric_type": {
                                    "type": "dict",
                                    "options": {
                                        "external": {"type": "bool"},
                                        "internal": {"type": "bool"},
                                        "rib_metric_as_external": {"type": "bool"},
                                        "rib_metric_as_internal": {"type": "bool"},
                                        "type_1": {"type": "bool"},
                                        "type_2": {"type": "bool"},
                                    },
                                },
                                "med": {
                                    "type": "dict",
                                    "options": {
                                        "value": {"type": "int"},
                                        "increment": {"type": "int"},
                                        "decrement": {"type": "int"},
                                        "igp_cost": {"type": "bool"},
                                        "max_reachable": {"type": "bool"},
                                        "parameter": {"type": "str"},
                                    },
                                },
                                "mpls": {"type": "str"},
                                "next_hop": {
                                    "type": "dict",
                                    "options": {"address": {"type": "str"}},
                                },
                                "origin": {
                                    "type": "dict",
                                    "options": {
                                        "egp": {"type": "bool"},
                                        "igp": {"type": "bool"},
                                        "rincomplete": {"type": "bool"},
                                    },
                                },
                                "ospf_metric": {"type": "int"},
                                "path_selection": {
                                    "type": "dict",
                                    "options": {
                                        "all": {"type": "bool"},
                                        "backup": {
                                            "type": "dict",
                                            "options": {
                                                "backup_decimal": {"type": "int"},
                                                "advertise": {"type": "bool"},
                                                "install": {"type": "bool"},
                                            },
                                        },
                                        "best_path": {"type": "bool"},
                                        "group_best": {"type": "bool"},
                                        "multiplath": {"type": "bool"},
                                    },
                                },
                                "path_color": {"type": "bool"},
                                "qos_group": {"type": "int"},
                                "rib_metric": {"type": "int"},
                                "rip_metric": {"type": "int"},
                                "rip_tag": {"type": "int"},
                                "rt_set": {"type": "int"},
                                "s_pmsi": {"type": "bool"},
                                "spf_priority": {
                                    "type": "dict",
                                    "options": {
                                        "critical": {"type": "bool"},
                                        "high": {"type": "bool"},
                                        "medium": {"type": "bool"},
                                    },
                                },
                                "static_p2mp_te": {"type": "str"},
                                "tag": {"type": "int"},
                                "traffic_index": {
                                    "type": "dict",
                                    "options": {
                                        "index_number": {"type": "int"},
                                        "ignore": {"type": "bool"},
                                    },
                                },
                                "upstream_core_tree": {
                                    "type": "dict",
                                    "options": {
                                        "ingress_replication": {"type": "bool"},
                                        "mldp": {"type": "bool"},
                                        "p2mp_te": {"type": "bool"},
                                        "sr_p2mp": {"type": "bool"},
                                    },
                                },
                                "vpn_distinguisher": {"type": "int"},
                                "weight": {"type": "int"},
                            },
                        },
                    },
                },
                "else_section": {
                    "type": "dict",
                    "options": {
                        "global": {
                            "type": "dict",
                            "options": {
                                "add": {
                                    "type": "dict",
                                    "options": {
                                        "eigrp_metric": {
                                            "type": "dict",
                                            "options": {
                                                "bandwidth": {"type": "int"},
                                                "delay": {"type": "int"},
                                                "reliability": {"type": "int"},
                                                "effective_bandwith": {
                                                    "type": "int",
                                                },
                                                "max_transmission": {
                                                    "type": "int",
                                                },
                                            },
                                        },
                                        "rip_metric": {"type": "int"},
                                    },
                                },
                                "apply": {
                                    "type": "list",
                                    "elements": "dict",
                                    "options": {
                                        "route_policy": {"type": "str"},
                                        "route_policy_input": {"type": "str"},
                                    },
                                },
                                "drop": {"type": "bool"},
                                "pass": {"type": "bool"},
                                "prepend": {
                                    "type": "dict",
                                    "options": {
                                        "number_of_times": {"type": "int"},
                                        "as_path": {"type": "int"},
                                        "most_recent": {"type": "bool"},
                                        "own_as": {"type": "bool"},
                                    },
                                },
                                "suppress_route": {"type": "bool"},
                                "unsuppress_route": {"type": "bool"},
                                "remove": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "entire_aspath": {"type": "bool"},
                                    },
                                },
                                "set": {
                                    "type": "dict",
                                    "options": {
                                        "administrative_distance": {"type": "int"},
                                        "local_preference": {
                                            "type": "list",
                                            "elements": "dict",
                                            "options": {
                                                "multiply": {"type": "bool"},
                                                "increment": {"type": "bool"},
                                                "decrement": {"type": "bool"},
                                                "metric_number": {
                                                    "type": "int",
                                                    "required": True,
                                                },
                                            },
                                        },
                                        "aigp_metric": {
                                            "type": "dict",
                                            "options": {
                                                "icrement": {"type": "bool"},
                                                "decrement": {"type": "bool"},
                                                "metric_number": {"type": "int"},
                                                "igp_cost": {"type": "bool"},
                                            },
                                        },
                                        "attribute_set": {"type": "str"},
                                        "c_multicast_routing": {
                                            "type": "dict",
                                            "options": {
                                                "bgp": {"type": "bool"},
                                                "pim": {"type": "bool"},
                                            },
                                        },
                                        "community": {
                                            "type": "dict",
                                            "options": {
                                                "community_name": {"type": "str"},
                                                "additive": {"type": "bool"},
                                            },
                                        },
                                        "core_tree": {
                                            "type": "dict",
                                            "options": {
                                                "ingress_replication": {
                                                    "type": "bool",
                                                },
                                                "ingress_replication_default": {
                                                    "type": "bool",
                                                },
                                                "ingress_replication_partitioned": {
                                                    "type": "bool",
                                                },
                                                "mldp": {"type": "bool"},
                                                "mldp_default": {"type": "bool"},
                                                "mldp_inband": {"type": "bool"},
                                                "mldp_partitioned_mp2mp": {
                                                    "type": "bool",
                                                },
                                                "mldp_partitioned_p2mp": {
                                                    "type": "bool",
                                                },
                                                "p2mp_te": {"type": "bool"},
                                                "p2mp_te_default": {
                                                    "type": "bool",
                                                },
                                                "p2mp_te_partitioned": {
                                                    "type": "bool",
                                                },
                                                "pim_default": {"type": "bool"},
                                                "sr_p2mp": {"type": "bool"},
                                            },
                                        },
                                        "dampening": {
                                            "type": "dict",
                                            "options": {
                                                "halflife": {"type": "int"},
                                                "max_suppress": {"type": "int"},
                                                "reuse": {"type": "int"},
                                                "suppress": {"type": "int"},
                                            },
                                        },
                                        "downstream_core_tree": {
                                            "type": "dict",
                                            "options": {
                                                "ingress_replication": {
                                                    "type": "bool",
                                                },
                                                "mldp": {"type": "bool"},
                                                "p2mp_te": {"type": "bool"},
                                                "sr_p2mp": {"type": "bool"},
                                            },
                                        },
                                        "eigrp_metric": {
                                            "type": "dict",
                                            "options": {
                                                "bandwidth": {"type": "int"},
                                                "delay": {"type": "int"},
                                                "reliability": {"type": "int"},
                                                "effective_bandwith": {
                                                    "type": "int",
                                                },
                                                "max_transmission": {
                                                    "type": "int",
                                                },
                                            },
                                        },
                                        "extcommunity": {
                                            "type": "dict",
                                            "options": {
                                                "soo": {"type": "str"},
                                                "rt": {"type": "str"},
                                                "additive": {"type": "bool"},
                                                "bandwidth": {"type": "str"},
                                                "color": {"type": "str"},
                                                "cost": {"type": "str"},
                                                "redirect_to_rt": {"type": "str"},
                                                "seg_nh": {"type": "str"},
                                            },
                                        },
                                        "fallback_vrf_lookup": {"type": "bool"},
                                        "flow_tag": {"type": "int"},
                                        "forward_class": {"type": "int"},
                                        "ip_precedence": {"type": "int"},
                                        "isis_metric": {"type": "int"},
                                        "label": {"type": "int"},
                                        "label_index": {"type": "int"},
                                        "label_mode": {
                                            "type": "dict",
                                            "options": {
                                                "per_ce": {"type": "bool"},
                                                "per_prefix": {"type": "bool"},
                                                "per_vrf": {"type": "bool"},
                                            },
                                        },
                                        "large_community": {"type": "str"},
                                        "level": {
                                            "type": "dict",
                                            "options": {
                                                "level_1": {"type": "bool"},
                                                "level_1_2": {"type": "bool"},
                                                "level_2": {"type": "bool"},
                                            },
                                        },
                                        "load_balance": {"type": "bool"},
                                        "lsm_root": {"type": "str"},
                                        "metric_type": {
                                            "type": "dict",
                                            "options": {
                                                "external": {"type": "bool"},
                                                "internal": {"type": "bool"},
                                                "rib_metric_as_external": {
                                                    "type": "bool",
                                                },
                                                "rib_metric_as_internal": {
                                                    "type": "bool",
                                                },
                                                "type_1": {"type": "bool"},
                                                "type_2": {"type": "bool"},
                                            },
                                        },
                                        "med": {
                                            "type": "dict",
                                            "options": {
                                                "value": {"type": "int"},
                                                "increment": {"type": "int"},
                                                "decrement": {"type": "int"},
                                                "igp_cost": {"type": "bool"},
                                                "max_reachable": {"type": "bool"},
                                                "parameter": {"type": "str"},
                                            },
                                        },
                                        "mpls": {"type": "str"},
                                        "next_hop": {
                                            "type": "dict",
                                            "options": {
                                                "address": {"type": "str"},
                                            },
                                        },
                                        "origin": {
                                            "type": "dict",
                                            "options": {
                                                "egp": {"type": "bool"},
                                                "igp": {"type": "bool"},
                                                "rincomplete": {"type": "bool"},
                                            },
                                        },
                                        "ospf_metric": {"type": "int"},
                                        "path_selection": {
                                            "type": "dict",
                                            "options": {
                                                "all": {"type": "bool"},
                                                "backup": {
                                                    "type": "dict",
                                                    "options": {
                                                        "backup_decimal": {
                                                            "type": "int",
                                                        },
                                                        "advertise": {
                                                            "type": "bool",
                                                        },
                                                        "install": {
                                                            "type": "bool",
                                                        },
                                                    },
                                                },
                                                "best_path": {"type": "bool"},
                                                "group_best": {"type": "bool"},
                                                "multiplath": {"type": "bool"},
                                            },
                                        },
                                        "path_color": {"type": "bool"},
                                        "qos_group": {"type": "int"},
                                        "rib_metric": {"type": "int"},
                                        "rip_metric": {"type": "int"},
                                        "rip_tag": {"type": "int"},
                                        "rt_set": {"type": "int"},
                                        "s_pmsi": {"type": "bool"},
                                        "spf_priority": {
                                            "type": "dict",
                                            "options": {
                                                "critical": {"type": "bool"},
                                                "high": {"type": "bool"},
                                                "medium": {"type": "bool"},
                                            },
                                        },
                                        "static_p2mp_te": {"type": "str"},
                                        "tag": {"type": "int"},
                                        "traffic_index": {
                                            "type": "dict",
                                            "options": {
                                                "index_number": {"type": "int"},
                                                "ignore": {"type": "bool"},
                                            },
                                        },
                                        "upstream_core_tree": {
                                            "type": "dict",
                                            "options": {
                                                "ingress_replication": {
                                                    "type": "bool",
                                                },
                                                "mldp": {"type": "bool"},
                                                "p2mp_te": {"type": "bool"},
                                                "sr_p2mp": {"type": "bool"},
                                            },
                                        },
                                        "vpn_distinguisher": {"type": "int"},
                                        "weight": {"type": "int"},
                                    },
                                },
                            },
                        },
                        "if_section": {
                            "type": "dict",
                            "options": {
                                "condition": {"type": "str"},
                                "add": {
                                    "type": "dict",
                                    "options": {
                                        "eigrp_metric": {
                                            "type": "dict",
                                            "options": {
                                                "bandwidth": {"type": "int"},
                                                "delay": {"type": "int"},
                                                "reliability": {"type": "int"},
                                                "effective_bandwith": {
                                                    "type": "int",
                                                },
                                                "max_transmission": {
                                                    "type": "int",
                                                },
                                            },
                                        },
                                        "rip_metric": {"type": "int"},
                                    },
                                },
                                "apply": {
                                    "type": "list",
                                    "elements": "dict",
                                    "options": {
                                        "route_policy": {"type": "str"},
                                        "route_policy_input": {"type": "str"},
                                    },
                                },
                                "drop": {"type": "bool"},
                                "pass": {"type": "bool"},
                                "prepend": {
                                    "type": "dict",
                                    "options": {
                                        "number_of_times": {"type": "int"},
                                        "as_path": {"type": "int"},
                                        "most_recent": {"type": "bool"},
                                        "own_as": {"type": "bool"},
                                    },
                                },
                                "suppress_route": {"type": "bool"},
                                "unsuppress_route": {"type": "bool"},
                                "remove": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "entire_aspath": {"type": "bool"},
                                    },
                                },
                                "set": {
                                    "type": "dict",
                                    "options": {
                                        "administrative_distance": {"type": "int"},
                                        "local_preference": {
                                            "type": "list",
                                            "elements": "dict",
                                            "options": {
                                                "multiply": {"type": "bool"},
                                                "increment": {"type": "bool"},
                                                "decrement": {"type": "bool"},
                                                "metric_number": {
                                                    "type": "int",
                                                    "required": True,
                                                },
                                            },
                                        },
                                        "aigp_metric": {
                                            "type": "dict",
                                            "options": {
                                                "icrement": {"type": "bool"},
                                                "decrement": {"type": "bool"},
                                                "metric_number": {"type": "int"},
                                                "igp_cost": {"type": "bool"},
                                            },
                                        },
                                        "attribute_set": {"type": "str"},
                                        "c_multicast_routing": {
                                            "type": "dict",
                                            "options": {
                                                "bgp": {"type": "bool"},
                                                "pim": {"type": "bool"},
                                            },
                                        },
                                        "community": {
                                            "type": "dict",
                                            "options": {
                                                "community_name": {"type": "str"},
                                                "additive": {"type": "bool"},
                                            },
                                        },
                                        "core_tree": {
                                            "type": "dict",
                                            "options": {
                                                "ingress_replication": {
                                                    "type": "bool",
                                                },
                                                "ingress_replication_default": {
                                                    "type": "bool",
                                                },
                                                "ingress_replication_partitioned": {
                                                    "type": "bool",
                                                },
                                                "mldp": {"type": "bool"},
                                                "mldp_default": {"type": "bool"},
                                                "mldp_inband": {"type": "bool"},
                                                "mldp_partitioned_mp2mp": {
                                                    "type": "bool",
                                                },
                                                "mldp_partitioned_p2mp": {
                                                    "type": "bool",
                                                },
                                                "p2mp_te": {"type": "bool"},
                                                "p2mp_te_default": {
                                                    "type": "bool",
                                                },
                                                "p2mp_te_partitioned": {
                                                    "type": "bool",
                                                },
                                                "pim_default": {"type": "bool"},
                                                "sr_p2mp": {"type": "bool"},
                                            },
                                        },
                                        "dampening": {
                                            "type": "dict",
                                            "options": {
                                                "halflife": {"type": "int"},
                                                "max_suppress": {"type": "int"},
                                                "reuse": {"type": "int"},
                                                "suppress": {"type": "int"},
                                            },
                                        },
                                        "downstream_core_tree": {
                                            "type": "dict",
                                            "options": {
                                                "ingress_replication": {
                                                    "type": "bool",
                                                },
                                                "mldp": {"type": "bool"},
                                                "p2mp_te": {"type": "bool"},
                                                "sr_p2mp": {"type": "bool"},
                                            },
                                        },
                                        "eigrp_metric": {
                                            "type": "dict",
                                            "options": {
                                                "bandwidth": {"type": "int"},
                                                "delay": {"type": "int"},
                                                "reliability": {"type": "int"},
                                                "effective_bandwith": {
                                                    "type": "int",
                                                },
                                                "max_transmission": {
                                                    "type": "int",
                                                },
                                            },
                                        },
                                        "extcommunity": {
                                            "type": "dict",
                                            "options": {
                                                "soo": {"type": "str"},
                                                "rt": {"type": "str"},
                                                "additive": {"type": "bool"},
                                                "bandwidth": {"type": "str"},
                                                "color": {"type": "str"},
                                                "cost": {"type": "str"},
                                                "redirect_to_rt": {"type": "str"},
                                                "seg_nh": {"type": "str"},
                                            },
                                        },
                                        "fallback_vrf_lookup": {"type": "bool"},
                                        "flow_tag": {"type": "int"},
                                        "forward_class": {"type": "int"},
                                        "ip_precedence": {"type": "int"},
                                        "isis_metric": {"type": "int"},
                                        "label": {"type": "int"},
                                        "label_index": {"type": "int"},
                                        "label_mode": {
                                            "type": "dict",
                                            "options": {
                                                "per_ce": {"type": "bool"},
                                                "per_prefix": {"type": "bool"},
                                                "per_vrf": {"type": "bool"},
                                            },
                                        },
                                        "large_community": {"type": "str"},
                                        "level": {
                                            "type": "dict",
                                            "options": {
                                                "level_1": {"type": "bool"},
                                                "level_1_2": {"type": "bool"},
                                                "level_2": {"type": "bool"},
                                            },
                                        },
                                        "load_balance": {"type": "bool"},
                                        "lsm_root": {"type": "str"},
                                        "metric_type": {
                                            "type": "dict",
                                            "options": {
                                                "external": {"type": "bool"},
                                                "internal": {"type": "bool"},
                                                "rib_metric_as_external": {
                                                    "type": "bool",
                                                },
                                                "rib_metric_as_internal": {
                                                    "type": "bool",
                                                },
                                                "type_1": {"type": "bool"},
                                                "type_2": {"type": "bool"},
                                            },
                                        },
                                        "med": {
                                            "type": "dict",
                                            "options": {
                                                "value": {"type": "int"},
                                                "increment": {"type": "int"},
                                                "decrement": {"type": "int"},
                                                "igp_cost": {"type": "bool"},
                                                "max_reachable": {"type": "bool"},
                                                "parameter": {"type": "str"},
                                            },
                                        },
                                        "mpls": {"type": "str"},
                                        "next_hop": {
                                            "type": "dict",
                                            "options": {
                                                "address": {"type": "str"},
                                            },
                                        },
                                        "origin": {
                                            "type": "dict",
                                            "options": {
                                                "egp": {"type": "bool"},
                                                "igp": {"type": "bool"},
                                                "rincomplete": {"type": "bool"},
                                            },
                                        },
                                        "ospf_metric": {"type": "int"},
                                        "path_selection": {
                                            "type": "dict",
                                            "options": {
                                                "all": {"type": "bool"},
                                                "backup": {
                                                    "type": "dict",
                                                    "options": {
                                                        "backup_decimal": {
                                                            "type": "int",
                                                        },
                                                        "advertise": {
                                                            "type": "bool",
                                                        },
                                                        "install": {
                                                            "type": "bool",
                                                        },
                                                    },
                                                },
                                                "best_path": {"type": "bool"},
                                                "group_best": {"type": "bool"},
                                                "multiplath": {"type": "bool"},
                                            },
                                        },
                                        "path_color": {"type": "bool"},
                                        "qos_group": {"type": "int"},
                                        "rib_metric": {"type": "int"},
                                        "rip_metric": {"type": "int"},
                                        "rip_tag": {"type": "int"},
                                        "rt_set": {"type": "int"},
                                        "s_pmsi": {"type": "bool"},
                                        "spf_priority": {
                                            "type": "dict",
                                            "options": {
                                                "critical": {"type": "bool"},
                                                "high": {"type": "bool"},
                                                "medium": {"type": "bool"},
                                            },
                                        },
                                        "static_p2mp_te": {"type": "str"},
                                        "tag": {"type": "int"},
                                        "traffic_index": {
                                            "type": "dict",
                                            "options": {
                                                "index_number": {"type": "int"},
                                                "ignore": {"type": "bool"},
                                            },
                                        },
                                        "upstream_core_tree": {
                                            "type": "dict",
                                            "options": {
                                                "ingress_replication": {
                                                    "type": "bool",
                                                },
                                                "mldp": {"type": "bool"},
                                                "p2mp_te": {"type": "bool"},
                                                "sr_p2mp": {"type": "bool"},
                                            },
                                        },
                                        "vpn_distinguisher": {"type": "int"},
                                        "weight": {"type": "int"},
                                    },
                                },
                            },
                        },
                        "elseif_section": {
                            "type": "list",
                            "elements": "dict",
                            "options": {
                                "condition": {"type": "str"},
                                "add": {
                                    "type": "dict",
                                    "options": {
                                        "eigrp_metric": {
                                            "type": "dict",
                                            "options": {
                                                "bandwidth": {"type": "int"},
                                                "delay": {"type": "int"},
                                                "reliability": {"type": "int"},
                                                "effective_bandwith": {
                                                    "type": "int",
                                                },
                                                "max_transmission": {
                                                    "type": "int",
                                                },
                                            },
                                        },
                                        "rip_metric": {"type": "int"},
                                    },
                                },
                                "apply": {
                                    "type": "list",
                                    "elements": "dict",
                                    "options": {
                                        "route_policy": {"type": "str"},
                                        "route_policy_input": {"type": "str"},
                                    },
                                },
                                "drop": {"type": "bool"},
                                "pass": {"type": "bool"},
                                "prepend": {
                                    "type": "dict",
                                    "options": {
                                        "number_of_times": {"type": "int"},
                                        "as_path": {"type": "int"},
                                        "most_recent": {"type": "bool"},
                                        "own_as": {"type": "bool"},
                                    },
                                },
                                "suppress_route": {"type": "bool"},
                                "unsuppress_route": {"type": "bool"},
                                "remove": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "entire_aspath": {"type": "bool"},
                                    },
                                },
                                "set": {
                                    "type": "dict",
                                    "options": {
                                        "administrative_distance": {"type": "int"},
                                        "local_preference": {
                                            "type": "list",
                                            "elements": "dict",
                                            "options": {
                                                "multiply": {"type": "bool"},
                                                "increment": {"type": "bool"},
                                                "decrement": {"type": "bool"},
                                                "metric_number": {
                                                    "type": "int",
                                                    "required": True,
                                                },
                                            },
                                        },
                                        "aigp_metric": {
                                            "type": "dict",
                                            "options": {
                                                "icrement": {"type": "bool"},
                                                "decrement": {"type": "bool"},
                                                "metric_number": {"type": "int"},
                                                "igp_cost": {"type": "bool"},
                                            },
                                        },
                                        "attribute_set": {"type": "str"},
                                        "c_multicast_routing": {
                                            "type": "dict",
                                            "options": {
                                                "bgp": {"type": "bool"},
                                                "pim": {"type": "bool"},
                                            },
                                        },
                                        "community": {
                                            "type": "dict",
                                            "options": {
                                                "community_name": {"type": "str"},
                                                "additive": {"type": "bool"},
                                            },
                                        },
                                        "core_tree": {
                                            "type": "dict",
                                            "options": {
                                                "ingress_replication": {
                                                    "type": "bool",
                                                },
                                                "ingress_replication_default": {
                                                    "type": "bool",
                                                },
                                                "ingress_replication_partitioned": {
                                                    "type": "bool",
                                                },
                                                "mldp": {"type": "bool"},
                                                "mldp_default": {"type": "bool"},
                                                "mldp_inband": {"type": "bool"},
                                                "mldp_partitioned_mp2mp": {
                                                    "type": "bool",
                                                },
                                                "mldp_partitioned_p2mp": {
                                                    "type": "bool",
                                                },
                                                "p2mp_te": {"type": "bool"},
                                                "p2mp_te_default": {
                                                    "type": "bool",
                                                },
                                                "p2mp_te_partitioned": {
                                                    "type": "bool",
                                                },
                                                "pim_default": {"type": "bool"},
                                                "sr_p2mp": {"type": "bool"},
                                            },
                                        },
                                        "dampening": {
                                            "type": "dict",
                                            "options": {
                                                "halflife": {"type": "int"},
                                                "max_suppress": {"type": "int"},
                                                "reuse": {"type": "int"},
                                                "suppress": {"type": "int"},
                                            },
                                        },
                                        "downstream_core_tree": {
                                            "type": "dict",
                                            "options": {
                                                "ingress_replication": {
                                                    "type": "bool",
                                                },
                                                "mldp": {"type": "bool"},
                                                "p2mp_te": {"type": "bool"},
                                                "sr_p2mp": {"type": "bool"},
                                            },
                                        },
                                        "eigrp_metric": {
                                            "type": "dict",
                                            "options": {
                                                "bandwidth": {"type": "int"},
                                                "delay": {"type": "int"},
                                                "reliability": {"type": "int"},
                                                "effective_bandwith": {
                                                    "type": "int",
                                                },
                                                "max_transmission": {
                                                    "type": "int",
                                                },
                                            },
                                        },
                                        "extcommunity": {
                                            "type": "dict",
                                            "options": {
                                                "soo": {"type": "str"},
                                                "rt": {"type": "str"},
                                                "additive": {"type": "bool"},
                                                "bandwidth": {"type": "str"},
                                                "color": {"type": "str"},
                                                "cost": {"type": "str"},
                                                "redirect_to_rt": {"type": "str"},
                                                "seg_nh": {"type": "str"},
                                            },
                                        },
                                        "fallback_vrf_lookup": {"type": "bool"},
                                        "flow_tag": {"type": "int"},
                                        "forward_class": {"type": "int"},
                                        "ip_precedence": {"type": "int"},
                                        "isis_metric": {"type": "int"},
                                        "label": {"type": "int"},
                                        "label_index": {"type": "int"},
                                        "label_mode": {
                                            "type": "dict",
                                            "options": {
                                                "per_ce": {"type": "bool"},
                                                "per_prefix": {"type": "bool"},
                                                "per_vrf": {"type": "bool"},
                                            },
                                        },
                                        "large_community": {"type": "str"},
                                        "level": {
                                            "type": "dict",
                                            "options": {
                                                "level_1": {"type": "bool"},
                                                "level_1_2": {"type": "bool"},
                                                "level_2": {"type": "bool"},
                                            },
                                        },
                                        "load_balance": {"type": "bool"},
                                        "lsm_root": {"type": "str"},
                                        "metric_type": {
                                            "type": "dict",
                                            "options": {
                                                "external": {"type": "bool"},
                                                "internal": {"type": "bool"},
                                                "rib_metric_as_external": {
                                                    "type": "bool",
                                                },
                                                "rib_metric_as_internal": {
                                                    "type": "bool",
                                                },
                                                "type_1": {"type": "bool"},
                                                "type_2": {"type": "bool"},
                                            },
                                        },
                                        "med": {
                                            "type": "dict",
                                            "options": {
                                                "value": {"type": "int"},
                                                "increment": {"type": "int"},
                                                "decrement": {"type": "int"},
                                                "igp_cost": {"type": "bool"},
                                                "max_reachable": {"type": "bool"},
                                                "parameter": {"type": "str"},  # like "$param1"
                                            },
                                        },
                                        "mpls": {"type": "str"},
                                        "next_hop": {
                                            "type": "dict",
                                            "options": {
                                                "address": {"type": "str"},
                                            },
                                        },
                                        "origin": {
                                            "type": "dict",
                                            "options": {
                                                "egp": {"type": "bool"},
                                                "igp": {"type": "bool"},
                                                "rincomplete": {"type": "bool"},
                                            },
                                        },
                                        "ospf_metric": {"type": "int"},
                                        "path_selection": {
                                            "type": "dict",
                                            "options": {
                                                "all": {"type": "bool"},
                                                "backup": {
                                                    "type": "dict",
                                                    "options": {
                                                        "backup_decimal": {
                                                            "type": "int",
                                                        },
                                                        "advertise": {
                                                            "type": "bool",
                                                        },
                                                        "install": {
                                                            "type": "bool",
                                                        },
                                                    },
                                                },
                                                "best_path": {"type": "bool"},
                                                "group_best": {"type": "bool"},
                                                "multiplath": {"type": "bool"},
                                            },
                                        },
                                        "path_color": {"type": "bool"},
                                        "qos_group": {"type": "int"},
                                        "rib_metric": {"type": "int"},
                                        "rip_metric": {"type": "int"},
                                        "rip_tag": {"type": "int"},
                                        "rt_set": {"type": "int"},
                                        "s_pmsi": {"type": "bool"},
                                        "spf_priority": {
                                            "type": "dict",
                                            "options": {
                                                "critical": {"type": "bool"},
                                                "high": {"type": "bool"},
                                                "medium": {"type": "bool"},
                                            },
                                        },
                                        "static_p2mp_te": {"type": "str"},
                                        "tag": {"type": "int"},
                                        "traffic_index": {
                                            "type": "dict",
                                            "options": {
                                                "index_number": {"type": "int"},
                                                "ignore": {"type": "bool"},
                                            },
                                        },
                                        "upstream_core_tree": {
                                            "type": "dict",
                                            "options": {
                                                "ingress_replication": {
                                                    "type": "bool",
                                                },
                                                "mldp": {"type": "bool"},
                                                "p2mp_te": {"type": "bool"},
                                                "sr_p2mp": {"type": "bool"},
                                            },
                                        },
                                        "vpn_distinguisher": {"type": "int"},
                                        "weight": {"type": "int"},
                                    },
                                },
                            },
                        },
                        "else_section": {
                            "type": "dict",
                            "options": {
                                "add": {
                                    "type": "dict",
                                    "options": {
                                        "eigrp_metric": {
                                            "type": "dict",
                                            "options": {
                                                "bandwidth": {"type": "int"},
                                                "delay": {"type": "int"},
                                                "reliability": {"type": "int"},
                                                "effective_bandwith": {
                                                    "type": "int",
                                                },
                                                "max_transmission": {
                                                    "type": "int",
                                                },
                                            },
                                        },
                                        "rip_metric": {"type": "int"},
                                    },
                                },
                                "apply": {
                                    "type": "list",
                                    "elements": "dict",
                                    "options": {
                                        "route_policy": {"type": "str"},
                                        "route_policy_input": {"type": "str"},
                                    },
                                },
                                "drop": {"type": "bool"},
                                "pass": {"type": "bool"},
                                "prepend": {
                                    "type": "dict",
                                    "options": {
                                        "number_of_times": {"type": "int"},
                                        "as_path": {"type": "int"},
                                        "most_recent": {"type": "bool"},
                                        "own_as": {"type": "bool"},
                                    },
                                },
                                "suppress_route": {"type": "bool"},
                                "unsuppress_route": {"type": "bool"},
                                "remove": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "entire_aspath": {"type": "bool"},
                                    },
                                },
                                "set": {
                                    "type": "dict",
                                    "options": {
                                        "administrative_distance": {"type": "int"},
                                        "local_preference": {
                                            "type": "list",
                                            "elements": "dict",
                                            "options": {
                                                "multiply": {"type": "bool"},
                                                "increment": {"type": "bool"},
                                                "decrement": {"type": "bool"},
                                                "metric_number": {
                                                    "type": "int",
                                                    "required": True,
                                                },
                                            },
                                        },
                                        "aigp_metric": {
                                            "type": "dict",
                                            "options": {
                                                "icrement": {"type": "bool"},
                                                "decrement": {"type": "bool"},
                                                "metric_number": {"type": "int"},
                                                "igp_cost": {"type": "bool"},
                                            },
                                        },
                                        "attribute_set": {"type": "str"},
                                        "c_multicast_routing": {
                                            "type": "dict",
                                            "options": {
                                                "bgp": {"type": "bool"},
                                                "pim": {"type": "bool"},
                                            },
                                        },
                                        "community": {
                                            "type": "dict",
                                            "options": {
                                                "community_name": {"type": "str"},
                                                "additive": {"type": "bool"},
                                            },
                                        },
                                        "core_tree": {
                                            "type": "dict",
                                            "options": {
                                                "ingress_replication": {
                                                    "type": "bool",
                                                },
                                                "ingress_replication_default": {
                                                    "type": "bool",
                                                },
                                                "ingress_replication_partitioned": {
                                                    "type": "bool",
                                                },
                                                "mldp": {"type": "bool"},
                                                "mldp_default": {"type": "bool"},
                                                "mldp_inband": {"type": "bool"},
                                                "mldp_partitioned_mp2mp": {
                                                    "type": "bool",
                                                },
                                                "mldp_partitioned_p2mp": {
                                                    "type": "bool",
                                                },
                                                "p2mp_te": {"type": "bool"},
                                                "p2mp_te_default": {
                                                    "type": "bool",
                                                },
                                                "p2mp_te_partitioned": {
                                                    "type": "bool",
                                                },
                                                "pim_default": {"type": "bool"},
                                                "sr_p2mp": {"type": "bool"},
                                            },
                                        },
                                        "dampening": {
                                            "type": "dict",
                                            "options": {
                                                "halflife": {"type": "int"},
                                                "max_suppress": {"type": "int"},
                                                "reuse": {"type": "int"},
                                                "suppress": {"type": "int"},
                                            },
                                        },
                                        "downstream_core_tree": {
                                            "type": "dict",
                                            "options": {
                                                "ingress_replication": {
                                                    "type": "bool",
                                                },
                                                "mldp": {"type": "bool"},
                                                "p2mp_te": {"type": "bool"},
                                                "sr_p2mp": {"type": "bool"},
                                            },
                                        },
                                        "eigrp_metric": {
                                            "type": "dict",
                                            "options": {
                                                "bandwidth": {"type": "int"},
                                                "delay": {"type": "int"},
                                                "reliability": {"type": "int"},
                                                "effective_bandwith": {
                                                    "type": "int",
                                                },
                                                "max_transmission": {
                                                    "type": "int",
                                                },
                                            },
                                        },
                                        "extcommunity": {
                                            "type": "dict",
                                            "options": {
                                                "soo": {"type": "str"},
                                                "rt": {"type": "str"},
                                                "additive": {"type": "bool"},
                                                "bandwidth": {"type": "str"},
                                                "color": {"type": "str"},
                                                "cost": {"type": "str"},
                                                "redirect_to_rt": {"type": "str"},
                                                "seg_nh": {"type": "str"},
                                            },
                                        },
                                        "fallback_vrf_lookup": {"type": "bool"},
                                        "flow_tag": {"type": "int"},
                                        "forward_class": {"type": "int"},
                                        "ip_precedence": {"type": "int"},
                                        "isis_metric": {"type": "int"},
                                        "label": {"type": "int"},
                                        "label_index": {"type": "int"},
                                        "label_mode": {
                                            "type": "dict",
                                            "options": {
                                                "per_ce": {"type": "bool"},
                                                "per_prefix": {"type": "bool"},
                                                "per_vrf": {"type": "bool"},
                                            },
                                        },
                                        "large_community": {"type": "str"},
                                        "level": {
                                            "type": "dict",
                                            "options": {
                                                "level_1": {"type": "bool"},
                                                "level_1_2": {"type": "bool"},
                                                "level_2": {"type": "bool"},
                                            },
                                        },
                                        "load_balance": {"type": "bool"},
                                        "lsm_root": {"type": "str"},
                                        "metric_type": {
                                            "type": "dict",
                                            "options": {
                                                "external": {"type": "bool"},
                                                "internal": {"type": "bool"},
                                                "rib_metric_as_external": {
                                                    "type": "bool",
                                                },
                                                "rib_metric_as_internal": {
                                                    "type": "bool",
                                                },
                                                "type_1": {"type": "bool"},
                                                "type_2": {"type": "bool"},
                                            },
                                        },
                                        "med": {
                                            "type": "dict",
                                            "options": {
                                                "value": {"type": "int"},
                                                "increment": {"type": "int"},
                                                "decrement": {"type": "int"},
                                                "igp_cost": {"type": "bool"},
                                                "max_reachable": {"type": "bool"},
                                                "parameter": {"type": "str"},  # like "$param1"
                                            },
                                        },
                                        "mpls": {"type": "str"},
                                        "next_hop": {
                                            "type": "dict",
                                            "options": {
                                                "address": {"type": "str"},
                                            },
                                        },
                                        "origin": {
                                            "type": "dict",
                                            "options": {
                                                "egp": {"type": "bool"},
                                                "igp": {"type": "bool"},
                                                "rincomplete": {"type": "bool"},
                                            },
                                        },
                                        "ospf_metric": {"type": "int"},
                                        "path_selection": {
                                            "type": "dict",
                                            "options": {
                                                "all": {"type": "bool"},
                                                "backup": {
                                                    "type": "dict",
                                                    "options": {
                                                        "backup_decimal": {
                                                            "type": "int",
                                                        },
                                                        "advertise": {
                                                            "type": "bool",
                                                        },
                                                        "install": {
                                                            "type": "bool",
                                                        },
                                                    },
                                                },
                                                "best_path": {"type": "bool"},
                                                "group_best": {"type": "bool"},
                                                "multiplath": {"type": "bool"},
                                            },
                                        },
                                        "path_color": {"type": "bool"},
                                        "qos_group": {"type": "int"},
                                        "rib_metric": {"type": "int"},
                                        "rip_metric": {"type": "int"},
                                        "rip_tag": {"type": "int"},
                                        "rt_set": {"type": "int"},
                                        "s_pmsi": {"type": "bool"},
                                        "spf_priority": {
                                            "type": "dict",
                                            "options": {
                                                "critical": {"type": "bool"},
                                                "high": {"type": "bool"},
                                                "medium": {"type": "bool"},
                                            },
                                        },
                                        "static_p2mp_te": {"type": "str"},
                                        "tag": {"type": "int"},
                                        "traffic_index": {
                                            "type": "dict",
                                            "options": {
                                                "index_number": {"type": "int"},
                                                "ignore": {"type": "bool"},
                                            },
                                        },
                                        "upstream_core_tree": {
                                            "type": "dict",
                                            "options": {
                                                "ingress_replication": {
                                                    "type": "bool",
                                                },
                                                "mldp": {"type": "bool"},
                                                "p2mp_te": {"type": "bool"},
                                                "sr_p2mp": {"type": "bool"},
                                            },
                                        },
                                        "vpn_distinguisher": {"type": "int"},
                                        "weight": {"type": "int"},
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        "running_config": {"type": "str"},
        "state": {
            "type": "str",
            "choices": [
                "merged",
                "replaced",
                "overridden",
                "purged",
                "rendered",
                "gathered",
                "parsed",
            ],
            "default": "merged",
        },
    }  # pylint: disable=C0301



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_networks_combine
short_description: Resource module for organizations _networks _combine
description:
  - Manage operation create of the resource organizations _networks _combine.
  - Combine multiple networks into a single network.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  enrollmentString:
    description: A unique identifier which can be used for device enrollment or easy
      access through the Meraki SM Registration page or the Self Service Portal. Please
      note that changing this field may cause existing bookmarks to break. All networks
      that are part of this combined network will have their enrollment string appended
      by '-network_type'. If left empty, all exisitng enrollment strings will be deleted.
    type: str
  name:
    description: The name of the combined network.
    type: str
  networkIds:
    description: A list of the network IDs that will be combined. If an ID of a combined
      network is included in this list, the other networks in the list will be grouped
      into that network.
    elements: str
    type: list
  organizationId:
    description: OrganizationId path parameter. Organization ID.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for organizations combineOrganizationNetworks
    description: Complete reference of the combineOrganizationNetworks API.
    link: https://developer.cisco.com/meraki/api-v1/#!combine-organization-networks
notes:
  - SDK Method used are
    organizations.Organizations.combine_organization_networks,
  - Paths used are
    post /organizations/{organizationId}/networks/combine,
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.organizations_networks_combine:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    enrollmentString: my-enrollment-string
    name: Long Island Office
    networkIds:
      - N_1234
      - N_5678
    organizationId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "resultingNetwork": {
        "id": "string",
        "organizationId": "string",
        "name": "string",
        "productTypes": [
          "string"
        ],
        "timeZone": "string",
        "tags": [
          "string"
        ],
        "enrollmentString": "string",
        "url": "string",
        "notes": "string",
        "isBoundToConfigTemplate": true
      }
    }
"""



# -*- coding: utf-8 -*-
# Copyright 2025 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

#############################################
#                WARNING                    #
#############################################
#
# This file is auto generated by the
# ansible.content_builder.
#
# Manually editing this file is not advised.
#
# To update the argspec make the desired changes
# in the documentation in the module file and re-run
# ansible.content_builder commenting out
# the path to external 'docstring' in build.yaml.
#
##############################################

"""
The arg spec for the ios_l2_interfaces module
"""


class L2_interfacesArgs(object):  # pylint: disable=R0903
    """The arg spec for the ios_l2_interfaces module"""

    argument_spec = {
        "config": {
            "type": "list",
            "elements": "dict",
            "options": {
                "name": {"type": "str", "required": True},
                "access": {
                    "type": "dict",
                    "options": {
                        "vlan": {"type": "int"},
                        "vlan_name": {"type": "str"},
                    },
                },
                "voice": {
                    "type": "dict",
                    "options": {
                        "vlan": {"type": "int"},
                        "vlan_tag": {
                            "choices": ["dot1p", "none", "untagged"],
                            "type": "str",
                        },
                        "vlan_name": {"type": "str"},
                    },
                },
                "trunk": {
                    "type": "dict",
                    "options": {
                        "allowed_vlans": {"type": "list", "elements": "str"},
                        "native_vlan": {"type": "int"},
                        "encapsulation": {
                            "choices": ["dot1q", "isl", "negotiate"],
                            "type": "str",
                        },
                        "pruning_vlans": {"type": "list", "elements": "str"},
                    },
                },
                "xconnect": {
                    "type": "dict",
                    "options": {
                        "address": {"type": "str"},
                        "vcid": {"type": "int"},
                        "encapsulation": {
                            "type": "str",
                            "choices": ["mpls", "l2tpv3"],
                        },
                        "manual": {"type": "bool"},
                        "pw_class": {"type": "str"},
                        "sequencing": {
                            "type": "str",
                            "choices": ["both", "receive", "transmit"],
                        },
                    },
                },
                "mode": {
                    "type": "str",
                    "choices": [
                        "access",
                        "trunk",
                        "dot1q_tunnel",
                        "dynamic",
                        "dynamic_auto",
                        "dynamic_desirable",
                        "private_vlan_host",
                        "private_vlan_promiscuous",
                        "private_vlan_trunk",
                    ],
                },
                "private_vlan": {
                    "type": "dict",
                    "options": {
                        "association": {"type": "bool"},
                        "host_association": {"type": "bool"},
                        "mapping": {"type": "bool"},
                        "host": {"type": "bool"},
                        "primary_range": {"type": "int"},
                        "secondary_range": {"type": "int"},
                        "add": {"type": "bool"},
                        "remove": {"type": "bool"},
                        "secondary_vlan_id": {"type": "str"},
                    },
                },
                "app_interface": {"type": "bool"},
                "nonegotiate": {"type": "bool"},
                "vepa": {"type": "bool"},
                "host": {"type": "bool"},
                "protected": {"type": "bool"},
                "encapsulation": {
                    "type": "dict",
                    "options": {
                        "type": {"type": "str"},
                        "vlan_id": {"type": "int"},
                    },
                },
                "block_options": {
                    "type": "dict",
                    "options": {
                        "multicast": {"type": "bool"},
                        "unicast": {"type": "bool"},
                    },
                },
                "spanning_tree": {
                    "type": "dict",
                    "options": {
                        "bpdufilter": {
                            "type": "dict",
                            "options": {
                                "enabled": {"type": "bool"},
                                "disabled": {"type": "bool"},
                            },
                        },
                        "bpduguard": {
                            "type": "dict",
                            "options": {
                                "enabled": {"type": "bool"},
                                "disabled": {"type": "bool"},
                            },
                        },
                        "cost": {"type": "int"},
                        "guard": {
                            "type": "dict",
                            "options": {
                                "loop": {"type": "bool"},
                                "none": {"type": "bool"},
                                "root": {"type": "bool"},
                            },
                        },
                        "link_type": {
                            "type": "dict",
                            "options": {
                                "point_to_point": {"type": "bool"},
                                "shared": {"type": "bool"},
                            },
                        },
                        "mst": {
                            "type": "dict",
                            "options": {
                                "instance_range": {"type": "str"},
                                "cost": {"type": "str"},
                                "port_priority": {"type": "int"},
                            },
                        },
                        "port_priority": {"type": "int"},
                        "portfast": {
                            "type": "dict",
                            "options": {
                                "disabled": {"type": "bool"},
                                "trunk": {"type": "bool"},
                            },
                        },
                        "rootguard": {"type": "bool"},
                        "vlan": {
                            "type": "dict",
                            "options": {
                                "vlan_range": {"type": "str"},
                                "cost": {"type": "str"},
                                "port_priority": {"type": "int"},
                            },
                        },
                    },
                },
            },
        },
        "running_config": {"type": "str"},
        "state": {
            "choices": [
                "merged",
                "replaced",
                "overridden",
                "deleted",
                "rendered",
                "gathered",
                "parsed",
            ],
            "default": "merged",
            "type": "str",
        },
    }  # pylint: disable=C0301



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2024, Akini Ross (@akinross) <akinross@cisco.com>

# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: ndo_tenant_mld_snooping_policy
short_description: Manage Multicast Listener Discovery (MLD) Snooping Policies in Tenant Policy Templates on Cisco Nexus Dashboard Orchestrator (NDO).
description:
- Manage MLD Snooping Policies in Tenant Policy Templates on Cisco Nexus Dashboard Orchestrator (NDO).
- This module is only supported on ND v3.1 (NDO v4.3) and later.
author:
- Akini Ross (@akinross)
options:
  template:
    description:
    - The name of the tenant template.
    type: str
    aliases: [ tenant_template ]
    required: true
  name:
    description:
    - The name of the MLD Snooping Policy.
    type: str
    aliases: [ mld_snooping_policy ]
  uuid:
    description:
    - The UUID of the MLD Snooping Policy.
    - This parameter is required when the MLD Snooping Policy O(name) needs to be updated.
    type: str
  description:
    description:
    - The description of the MLD Snooping Policy.
    - Providing an empty string will remove the O(description="") from the MLD Snooping Policy.
    type: str
  admin_state:
    description:
    - The administrative state of the MLD Snooping Policy.
    - Defaults to C(disabled) when unset during creation.
    type: str
    choices: [ enabled, disabled ]
  fast_leave_control:
    description:
    - The fast leave control of the MLD Snooping Policy.
    - Defaults to C(disabled) when unset during creation.
    type: str
    choices: [ enabled, disabled ]
  querier_control:
    description:
    - The querier control of the MLD Snooping Policy.
    - Defaults to C(disabled) when unset during creation.
    type: str
    choices: [ enabled, disabled ]
  querier_version:
    description:
    - The querier version of the MLD Snooping Policy.
    - Defaults to C(v2) when unset during creation.
    type: str
    choices: [ v1, v2 ]
  query_interval:
    description:
    - The query interval of the MLD Snooping Policy in seconds.
    - Defaults to 125 when unset during creation.
    - The value must be between 1 and 18000.
    type: int
  query_response_interval:
    description:
    - The query response interval of the MLD Snooping Policy in seconds.
    - Defaults to 10 when unset during creation.
    - The value must be between 1 and 25.
    type: int
  last_member_query_interval:
    description:
    - The last member query interval of the MLD Snooping Policy in seconds.
    - Defaults to 1 when unset during creation.
    - The value must be between 1 and 25.
    type: int
  start_query_interval:
    description:
    - The start query interval of the MLD Snooping Policy in seconds.
    - Defaults to 31 when unset during creation.
    - The value must be between 1 and 18000.
    type: int
  start_query_count:
    description:
    - The start query count of the MLD Snooping Policy.
    - Defaults to 2 when unset during creation.
    - The value must be between 1 and 10.
    type: int
  state:
    description:
    - Use C(absent) for removing.
    - Use C(query) for listing an object or multiple objects.
    - Use C(present) for creating or updating.
    type: str
    choices: [ absent, query, present ]
    default: query
notes:
- The O(template) must exist before using this module in your playbook.
  Use M(cisco.mso.ndo_template) to create the Tenant template.
seealso:
- module: cisco.mso.ndo_template
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Create a new MLD Snooping Policy object
  cisco.mso.ndo_tenant_mld_snooping_policy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: tenant_template
    name: mld_snooping_policy_1
    admin_state: enabled
    fast_leave_control: enabled
    querier_control: enabled
    querier_version: v1
    query_interval: 100
    query_response_interval: 5
    last_member_query_interval: 2
    start_query_interval: 25
    start_query_count: 1
    state: present
  register: mld_snooping_policy_1

- name: Update a MLD Snooping Policy object name with UUID
  cisco.mso.ndo_tenant_mld_snooping_policy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: tenant_template
    name: mld_snooping_policy_2
    uuid: "{{ mld_snooping_policy_1.current.uuid }}"
    state: present

- name: Query a MLD Snooping Policy object with name
  cisco.mso.ndo_tenant_mld_snooping_policy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: tenant_template
    name: mld_snooping_policy_1
    state: query
  register: query_name

- name: Query a MLD Snooping Policy object with UUID
  cisco.mso.ndo_tenant_mld_snooping_policy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: tenant_template
    uuid: "{{ mld_snooping_policy_1.current.uuid }}"
    state: query
  register: query_uuid

- name: Query all MLD Snooping Policy objects in a Tenant Template
  cisco.mso.ndo_tenant_mld_snooping_policy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: tenant_template
    state: query
  register: query_all

- name: Delete a MLD Snooping Policy object with name
  cisco.mso.ndo_tenant_mld_snooping_policy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: tenant_template
    name: mld_snooping_policy_1
    state: absent

- name: Delete a MLD Snooping Policy object with UUID
  cisco.mso.ndo_tenant_mld_snooping_policy:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: tenant_template
    uuid: "{{ mld_snooping_policy_1.current.uuid }}"
    state: absent
"""

RETURN = r"""
"""

import copy
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec
from ansible_collections.cisco.mso.plugins.module_utils.template import MSOTemplate, KVPair
from ansible_collections.cisco.mso.plugins.module_utils.constants import ENABLED_OR_DISABLED_TO_BOOL_STRING_MAP
from ansible_collections.cisco.mso.plugins.module_utils.utils import append_update_ops_data


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        template=dict(type="str", required=True, aliases=["tenant_template"]),
        name=dict(type="str", aliases=["mld_snooping_policy"]),
        uuid=dict(type="str"),
        description=dict(type="str"),
        admin_state=dict(type="str", choices=["enabled", "disabled"]),
        fast_leave_control=dict(type="str", choices=["enabled", "disabled"]),
        querier_control=dict(type="str", choices=["enabled", "disabled"]),
        querier_version=dict(type="str", choices=["v1", "v2"]),
        query_interval=dict(type="int"),
        query_response_interval=dict(type="int"),
        last_member_query_interval=dict(type="int"),
        start_query_interval=dict(type="int"),
        start_query_count=dict(type="int"),
        state=dict(type="str", default="query", choices=["absent", "query", "present"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["name", "uuid"], True],
            ["state", "present", ["name", "uuid"], True],
        ],
    )

    mso = MSOModule(module)

    template = module.params.get("template")
    name = module.params.get("name")
    uuid = module.params.get("uuid")
    description = module.params.get("description")
    admin_state = module.params.get("admin_state")
    fast_leave_control = ENABLED_OR_DISABLED_TO_BOOL_STRING_MAP.get(module.params.get("fast_leave_control"))
    querier_control = ENABLED_OR_DISABLED_TO_BOOL_STRING_MAP.get(module.params.get("querier_control"))
    querier_version = module.params.get("querier_version")
    query_interval = module.params.get("query_interval")
    query_response_interval = module.params.get("query_response_interval")
    last_member_query_interval = module.params.get("last_member_query_interval")
    start_query_interval = module.params.get("start_query_interval")
    start_query_count = module.params.get("start_query_count")
    state = module.params.get("state")

    template_object = MSOTemplate(mso, "tenant", template)
    template_object.validate_template("tenantPolicy")

    mld_snooping_policies = template_object.template.get("tenantPolicyTemplate", {}).get("template", {}).get("mldSnoopPolicies", [])
    object_description = "MLD Snooping Policy"
    mld_snooping_policy_attrs_path = None
    match = None

    if state in ["query", "absent"] and mld_snooping_policies == []:
        mso.exit_json()
    elif state == "query" and not (name or uuid):
        mso.existing = mld_snooping_policies
    elif mld_snooping_policies and (name or uuid):
        match = template_object.get_object_by_key_value_pairs(
            object_description, mld_snooping_policies, [KVPair("uuid", uuid) if uuid else KVPair("name", name)]
        )
        if match:
            mld_snooping_policy_attrs_path = "/tenantPolicyTemplate/template/mldSnoopPolicies/{0}".format(match.index)
            mso.existing = mso.previous = copy.deepcopy(match.details)

    ops = []

    if state == "present":
        if uuid and not mso.existing:
            mso.fail_json(msg="{0} with the UUID: '{1}' not found".format(object_description, uuid))

        mso_values = dict(
            name=name,
            description=description,
            enableAdminState=admin_state,
            enableFastLeaveControl=fast_leave_control,
            enableQuerierControl=querier_control,
            mldQuerierVersion=querier_version,
            queryInterval=query_interval,
            queryResponseInterval=query_response_interval,
            lastMemberQueryInterval=last_member_query_interval,
            startQueryInterval=start_query_interval,
            startQueryCount=start_query_count,
        )

        if mso.existing and match:
            append_update_ops_data(ops, match.details, mld_snooping_policy_attrs_path, mso_values)
            mso.sanitize(match.details, collate=True)
        else:
            mso.sanitize(mso_values)
            ops.append(dict(op="add", path="/tenantPolicyTemplate/template/mldSnoopPolicies/-", value=mso.sent))

    elif state == "absent":
        if mso.existing and match:
            ops.append(dict(op="remove", path=mld_snooping_policy_attrs_path))

    if not module.check_mode and ops:
        response_object = mso.request(template_object.template_path, method="PATCH", data=ops)
        mld_snooping_policies = response_object.get("tenantPolicyTemplate", {}).get("template", {}).get("mldSnoopPolicies", [])
        match = template_object.get_object_by_key_value_pairs(
            object_description, mld_snooping_policies, [KVPair("uuid", uuid) if uuid else KVPair("name", name)]
        )
        if match:
            mso.existing = match.details  # When the state is present
        else:
            mso.existing = {}  # When the state is absent
    elif module.check_mode and state != "query":  # When the state is present/absent with check mode
        mso.existing = mso.proposed if state == "present" else {}

    mso.exit_json()


if __name__ == "__main__":
    main()
